/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "music.h"

namespace mysdl
{
	namespace audio
	{
		MusicStream::MusicStream(MusicManager *owner,StreamedSample *sample) :
			owner(owner),sample(sample),loader(0),bufferSize(0),fileComplete(false),loop(false),buffers()
		{
			bufferSize=sample->bufferSize;
			loop=sample->properties.loop;

			for(int i=0; i<BUFFERS_PER_STREAM; i++)
			{
				buffers[i]=0;
			}

			try
			{
				loader=getFileLoader(sample->filename);
			}
			catch(std::string ex)
			{
				if(loader)
				{
					delete loader;
					loader=0;
				}

				throw std::string("Error streaming music \"")+sample->filename+std::string("\": ")+ex;
			}

			alGenBuffers(BUFFERS_PER_STREAM,buffers);
		}

		MusicStream::~MusicStream()
		{
			if(buffers[0])
			{
				alDeleteBuffers(BUFFERS_PER_STREAM,buffers);

				for(int i=0; i<BUFFERS_PER_STREAM; i++)
				{
					buffers[i]=0;
				}
			}

			if(loader)
			{
				delete loader;
				loader=0;
			}
		}

		bool MusicStream::hasBuffer(ALuint buffer)
		{
			/*for(int i=0; i<BUFFERS_PER_STREAM; i++)
			{
				if(buffer==buffers[i])
				{
					return true;
				}
			}

			return false;*/
			return((buffer==buffers[0])||(buffer==buffers[1]));
		}

		bool MusicStream::stream(ALuint buffer,bool newStream)
		{
// 			if(!newStream)
// 			{
// 				return buffer;
// 			}

			ALsizei size=bufferSize;
			char *data=loader->read(&size,loop);

			if(size)
			{
				alBufferData(buffer,loader->getFormat(),data,size,loader->getFrequency());
				return true;
			}
			else
			{
				fileComplete=true;
				return false;
			}
		}

		int MusicManager_run(void *manager)
		{
			return(((MusicManager *)manager)->run());
		}

		MusicManager::MusicManager() :
			stopped(true),
			sourceName(0),
			queue(0),
			currentSound(0),
			running(true),
			thread(0),
			mutex(0),
			volume(1.0),
			currentVolume(1.0),
			mode(mmmNormal),
			modeParam(0),
			currentStream(0),
			nextStream(0)
		{
			alGenSources(1,&sourceName);

			mutex=SDL_CreateMutex();
			thread=SDL_CreateThread(MusicManager_run,this);
		}

		MusicManager::~MusicManager()
		{
			stop();

			if(thread)
			{
				terminate();
			}

			if(mutex)
			{
				SDL_DestroyMutex(mutex);
				mutex=0;
			}

			alDeleteSources(1,&sourceName);
		}

		int MusicManager::run()
		{
			Uint32 lastTicks=SDL_GetTicks();

			while(running)
			{
				Uint32 currentTicks=SDL_GetTicks();
				perform(currentTicks-lastTicks);
				lastTicks=currentTicks;
				SDL_Delay(100);//250);
			}

			return 0;
		}

		int MusicManager::terminate()
		{
			int result;

			running=false;
			SDL_WaitThread(thread,&result);

			return result;
		}

		void MusicManager::skipToNext()
		{
			stop();
			play();
		}

		void MusicManager::skipToLast()
		{
			while(queue.size()>1)
			{
				queue.pop_front();
			}

			stop();
			play();
		}

		// synchronized
		void MusicManager::addAsNext(StreamedSample *sample)
		{
			SDL_mutexP(mutex);
			queue.push_front(sample);
			SDL_mutexV(mutex);
		}

		// synchronized
		void MusicManager::addAsLast(StreamedSample *sample)
		{
			SDL_mutexP(mutex);
			queue.push_back(sample);
			SDL_mutexV(mutex);
		}

		void MusicManager::clearQueue()
		{
			queue.clear();
		}

		bool MusicManager::isPlaying()
		{
			return !stopped;
		}

		void MusicManager::setVolume(ALfloat volume)
		{
			this->volume=volume;
		}

		void MusicManager::setCurrentVolume(ALfloat volume)
		{
			this->currentVolume=volume;
			alSourcef(sourceName,AL_GAIN,volume);
		}

		void MusicManager::fadeOut(Uint32 ticks)
		{
			mode=mmmFadeOut;
			modeParam=((ALfloat)1.0)/((ALfloat)ticks);
		}

		// synchronized
		void MusicManager::perform(Uint32 ticks)
		{
			bool doStop=false;

			SDL_mutexP(mutex);

			if(mode==mmmFadeOut)
			{
				currentVolume-=modeParam*(ALfloat)ticks;

				if(currentVolume<=0)
				{
					currentVolume=volume;
					doStop=true;
				}
				else
				{
					setCurrentVolume(currentVolume);
				}
			}

			if(!doStop)
			{
				stream();
			}

			SDL_mutexV(mutex);

			if(doStop)
			{
				stop();
			}
		}

		// synchronized
		void MusicManager::stop()
		{
			SDL_mutexP(mutex);

			alSourceStop(sourceName);
			alSourcei(sourceName,AL_BUFFER,AL_NONE);

			if(nextStream)
			{
				delete nextStream;
				nextStream=0;
			}

			if(currentStream)
			{
				delete currentStream;
				currentStream=0;
			}

			stopped=true;

			SDL_mutexV(mutex);
		}

		void MusicManager::stream()
		{
			bool playing=false;
			ALint processed;

			//for(alGetSourcei(sourceName,AL_BUFFERS_PROCESSED,&processed); processed>0; processed--)
			alGetSourcei(sourceName,AL_BUFFERS_PROCESSED,&processed);
			if(processed>0)
			{
				ALuint buffer;

				alSourceUnqueueBuffers(sourceName,1,&buffer);

				if(currentStream->hasBuffer(buffer))
				{
					if((!(currentStream->fileComplete))||(!currentStreamComplete))
					{
						currentStream->stream(buffer,false);

						if(!currentStream->fileComplete)
						{
							alSourceQueueBuffers(sourceName,1,&buffer);
							playing=true;
//  							printf("Queuing next buffer of current stream.\n");fflush(stdout);
						}
						else
						{
							currentStreamComplete=true;

							if(queue.size()>0)
							{
								StreamedSample *sample=queue.front();
								queue.pop_front();
								nextStream=new MusicStream(this,sample);
								int count=0;

								for(int i=0; i<BUFFERS_PER_STREAM; i++)
								{
									if(nextStream->stream(nextStream->buffers[i],true))
									{
										count++;
										playing=true;
									}
								}

								alSourceQueueBuffers(sourceName,count,nextStream->buffers);
//  								printf("Next music segment starts with %d buffers.\n",count);fflush(stdout);
							}
						}
					}
				}
				else
				{
					delete currentStream;
					currentStream=nextStream;
					nextStream=0;
					currentStreamComplete=false;

					if(currentStream)
					{
						currentStream->stream(buffer,false);
						alSourceQueueBuffers(sourceName,1,&buffer);
						playing=true;
//  						printf("Queuing next buffer (old stream just finished).\n");fflush(stdout);
					}
				}
			}
			
			if(playing)
			{
				ALint hasStopped;
				alGetSourcei(sourceName,AL_SOURCE_STATE,&hasStopped);

				if(hasStopped==AL_STOPPED)
				{
					alSourcePlay(sourceName);
				}
			}
		}

		// synchronized
		void MusicManager::play()
		{
			SDL_mutexP(mutex); // TODO: move all this to the music thread

			if(queue.size()>0)
			{
				currentStreamComplete=false;
				mode=mmmNormal;
				stopped=false;
				currentVolume=volume;
 				StreamedSample *sample=queue.front();
 				queue.pop_front();
				currentStream=new MusicStream(this,sample);

				int count=0;

				for(int i=0; i<BUFFERS_PER_STREAM; i++)
				{
					if(currentStream->stream(currentStream->buffers[i],true))
					{
						count++;
					}
				}

				alSourceQueueBuffers(sourceName,count,currentStream->buffers);

				ALfloat pos[]={0.0,0.0,0.0};
				
				alSourcef(sourceName,AL_GAIN,volume);
				alSourcef(sourceName,AL_PITCH,1.0);
				alSourcef(sourceName,AL_ROLLOFF_FACTOR,0.0);
				alSourcefv(sourceName,AL_POSITION,pos);
				alSourcei(sourceName,AL_LOOPING,AL_FALSE);
				alSourcePlay(sourceName);
				
//  				printf("Music starts with %d buffers.\n",count);fflush(stdout);
			}

			SDL_mutexV(mutex);
		}
	};
};
