/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "loaders.h"
#include "resources.h"

namespace mysdl
{
	namespace audio
	{
		static const std::string XML_SOUNDS_SAMPLES="samples";
		static const std::string XML_SOUNDS_DIR="dir";
		static const std::string XML_SOUNDS_SAMPLE="sample";
		static const std::string XML_SOUNDS_STREAM="stream";
		static const std::string XML_SOUNDS_NAME="name";
		static const std::string XML_SOUNDS_FILENAME="filename";
		static const std::string XML_SOUNDS_VOLUME="volume";
		static const std::string XML_SOUNDS_PITCH="pitch";
		static const std::string XML_SOUNDS_ROLLOFF="rolloff";
		static const std::string XML_SOUNDS_LOOP="loop";
		static const std::string XML_SOUNDS_BUFFERSIZE="bufferSize";
		static const std::string XML_SOUNDS_PRIORITY="priority";

		//static const long DEFAULT_BUFFERSIZE=4096;//*10;

		SoundProperties::SoundProperties() :
			position(),
			volume(1.0),
			rolloff(1.0),
			pitch(1.0),
			loop(false)
		{
		}

		void SoundProperties::setStereoPosition(GLfloat stereoPosition)
		{
			if(stereoPosition<=-1)
			{
				position.x=-1;
				position.z=0;
			}
			else if(stereoPosition>=+1)
			{
				position.x=+1;
				position.z=0;
			}
			else if(stereoPosition<0)
			{
				position.x=stereoPosition;
				position.z=1.0+stereoPosition;//-stereoPosition-1.0;
			}
			else
			{
				position.x=stereoPosition;
				position.z=1.0-stereoPosition;
			}
		}

		Sample::Sample() : NamedResource<SampleBank>(),
			filename(),/*loop(false),*/priority(0),properties()
		{
		}

		Sample::Sample(const xml::Node *root,const std::string &baseDir) : NamedResource<SampleBank>(),
			filename(),/*loop(false),*/priority(0),properties()
		{
			getFromNode(root,baseDir);
		}

		Sample::~Sample()
		{
			release();
		}

		void Sample::getFromNode(const xml::Node *node,const std::string &baseDir)
		{
			const std::string *newName=node->getAttribute(XML_SOUNDS_NAME);
			const std::string *newFilename=node->getAttribute(XML_SOUNDS_FILENAME);

			if(!newName)
			{
				debug::warning(dbgImages,"No name specified for sound. Using filename.");
				newName=newFilename;
			}

			if(!newFilename)
			{
				throw std::string("No filename specified for sound \"")+(*newName)+std::string("\".");
			}

			name=*newName;
			filename=baseDir+*newFilename;

//			loop=node->getAttributeBool(XML_SOUNDS_LOOP,loop);

			properties.volume=node->getAttributeFloat(XML_SOUNDS_VOLUME,properties.volume);
			properties.pitch=node->getAttributeFloat(XML_SOUNDS_PITCH,properties.pitch);
			properties.rolloff=node->getAttributeFloat(XML_SOUNDS_ROLLOFF,properties.rolloff);
			properties.loop=node->getAttributeBool(XML_SOUNDS_LOOP,properties.loop);

			priority=(int)(node->getAttributeLong(XML_SOUNDS_PRIORITY,priority));
		}

		void Sample::release()
		{
		}

		void Sample::reload()
		{
		}

		void Sample::refresh()
		{
		}

		StreamedSample::StreamedSample() :
			Sample(),bufferSize(DEF_BUFFER_SIZE)
		{
		}

		StreamedSample::StreamedSample(const xml::Node *root,const std::string &baseDir) :
			Sample(),bufferSize(DEF_BUFFER_SIZE)
		{
			getFromNode(root,baseDir);
		}

		void StreamedSample::getFromNode(const xml::Node *node,const std::string &baseDir)
		{
			Sample::getFromNode(node,baseDir);

			bufferSize=node->getAttributeLong(XML_SOUNDS_BUFFERSIZE,bufferSize);
		}

		CachedSample::CachedSample() : Sample()
		{
			bufferName=0;
			format=0;
			size=0;
			frequency=0;
		}

		CachedSample::CachedSample(const xml::Node *root,const std::string &baseDir) : Sample()
		{
			bufferName=0;
			format=0;
			size=0;
			frequency=0;

			getFromNode(root,baseDir);
		}

		CachedSample::~CachedSample()
		{
			release();
		}

		void CachedSample::getFromNode(const xml::Node *node,const std::string &baseDir)
		{
			Sample::getFromNode(node,baseDir);
		}

		void CachedSample::release()
		{
			if(bufferName)
			{
				alDeleteBuffers(1,&bufferName);
				bufferName=0;
			}
		}

		void CachedSample::load()
		{
			if(enabled)
			{
				alGetError();
				alGenBuffers(1,&bufferName);

				if(!bufferName)
				{
					ALenum error=alGetError();
					throw std::string("Could not create OpenAL buffer for \"")+name+std::string("\": ")+getALErrorString(error)+std::string(".");
				}

				FileLoader *loader=0;

				try
				{
					loader=getFileLoader(filename);

					char *data=loader->read();
					alBufferData(bufferName,loader->getFormat(),data,loader->getSize(),loader->getFrequency());

					format=loader->getFormat();
					frequency=loader->getFrequency();
					size=loader->getSize();
					
					delete loader;
					loader=0;
				}
				catch(std::string ex)
				{
					if(loader)
					{
						delete loader;
						loader=0;
					}

					throw std::string("Error loading file \"")+filename+std::string("\": ")+ex;
				}

				debug::debug(dbgAudio,"Loaded \"%s\" as sound \"%s\" (#%d), format: %s, size: %d, frequency: %d.",filename.c_str(),name.c_str(),(int)bufferName,(format==AL_FORMAT_MONO16 ? "mono" : "stereo"),(int)size,(int)frequency);
			}
		}

		void CachedSample::reload()
		{
			release();
			load();
		}

		void CachedSample::refresh()
		{
			if(!bufferName)
			{
				debug::debug(dbgAudio,"Refreshing sound \"%s\"...",name.c_str());
				reload();
			}
		}

		void SampleBank::reload()
		{
			for(iterator itr=begin(); itr!=end(); itr++)
			{
				itr->reload();
			}
		}

		void SampleBank::refresh()
		{
			for(iterator itr=begin(); itr!=end(); itr++)
			{
				itr->refresh();
			}
		}

		/**
		 * Releases all sounds in the bank.
		 */
		void SampleBank::release()
		{
			for(iterator itr=begin(); itr!=end(); itr++)
			{
				itr->release();
			}
		}

		void SampleBank::getFromNode(const xml::Node *root,const std::string &baseDir)
		{
			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				const xml::Node *node=*itr;

				if(node->name==XML_SOUNDS_SAMPLE)
				{
					Sample *sound=new CachedSample(node,baseDir);
					(*this)+=sound;

					debug::debug(dbgAudio,"Added cached sample \"%s\" (%p).",sound->name.c_str(),sound);
				}
				else if(node->name==XML_SOUNDS_STREAM)
				{
					Sample *sound=new StreamedSample(node,baseDir);
					(*this)+=sound;

					debug::debug(dbgAudio,"Added streamed sample \"%s\" (%p).",sound->name.c_str(),sound);
				}
			}
		}

		void SampleBank::load(const std::string &filename)
		{
			// --- Determine the path of the file (sound filenames are ---
			// --- considered to be relative to the path of the file   ---
			// --- defining the bank).                                 ---
			std::string::size_type i=filename.find_first_of('/',0);
			std::string dir;

			if(i!=std::string::npos)
			{
				std::string::size_type j;

				while((j=filename.find_first_of('/',i+1))!=std::string::npos)
				{
					i=j;
				}

				dir=filename.substr(0,i+1);
			}
			else
			{
				dir="";
			}

			xml::Node *root=xml::loadXML(filename.c_str());

			if((root)&&(root->name==XML_SOUNDS_SAMPLES))
			{
				const std::string *subDir=0;

				if((subDir=root->getAttribute(XML_SOUNDS_DIR)))
				{
					dir+=*subDir;
				}

				getFromNode(root,dir);
			}
		}

		SampleBank::~SampleBank()
		{
			mysdl::debug::debug(dbgAudio,"Freeing sample bank %p.",this);
		}
	};
};
