/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <iostream>
#include <signal.h>

#include "debug.h"
#include "mysdl.h"
#include "xml.h"

namespace mysdl
{
	namespace debug
	{
		const char *const XML_DEBUG="debug";
		const char *const XML_GROUP="group";
		const char *const XML_NAME ="name";
		const char *const XML_LEVEL="level";

		static const char *LEVEL_NAME[]={"none","critical","error","warning","info","debug"};

		std::string previousMessage=std::string("");

		/** A list/stack of strings used to store debug messages. These are usually
		*  printed to stderr from MySDL's default signal handler. */
  	std::list<std::string> messages=std::list<std::string>();

		Groups groups=Groups();

		DebugLevel debugLevel(const char *levelName)
		{
			DebugLevel level=dlNone;

			for(int l=dlNone; l<=dlDebug; l++)
			{
				if(strcmp(levelName,LEVEL_NAME[l])==0)
				{
					level=(DebugLevel)l; //TODO: ugly cast?
				}
			}

			return(level);
		}

		void Groups::getSettingsFromNode(const xml::Node *root)
		{
			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				xml::Node *node=(*itr);

				if(node->name==XML_GROUP)
				{
					const std::string *attrName=node->getAttribute(XML_NAME);
					const std::string *attrLevel=node->getAttribute(XML_LEVEL);

					if((attrName)&&(attrLevel))
					{
						for(Groups::iterator itr=begin(); itr!=end(); itr++)
						{
							if(itr->second.name==*attrName)
							{
								itr->second.level=debugLevel(attrLevel->c_str());
							}
						}
					}
				}
			}
		}

		void Groups::loadSettings(const char *filename)
		{
			//addMessage("Groups::loadSettings()");

			xml::Node *root=xml::loadXML(filename);

			if(root)
			{
				if(root->name==XML_DEBUG)
				{
					getSettingsFromNode(root);
				}

				delete root;
				root=0;
			}

			//removeMessage();
		}

		Group::Group() : name("undefined"),level(DEF_DEBUG_LEVEL)
		{
		}

		Group::Group(const std::string name,DebugLevel level) : name(name),level(level)
		{
		}

		/**
		 * Prints a debug message, if the specified debug level includes the debug level
		 * of the specified debug group.
		 *
		 * @param level Current debug level.
		 * @param group Debug group for which the message will be printed.
		 * @param fmt   Format string. See printf() for details.
		 * @param ap    Arbitrary number of parameters. See printf() for details.
		 */
		void vmessagef(DebugLevel level,int group,const char *fmt,/*const*/ va_list *ap)
		{
			//addMessage("messagef()");

			if((fmt)&&(groups[group].level>=level))
			{
				printf("[%s] (%s) ",groups[group].name.c_str(),LEVEL_NAME[level]);
				vprintf(fmt,(*ap));
				printf("\n");
				fflush(stdout);
			}

			//removeMessage();
		}

		#ifdef MYSDL_DEBUG_CRITICAL
		void critical(int group,const char *fmt,...)
		{
			va_list ap;

			va_start(ap,fmt);
			vmessagef(dlCritical,group,fmt,&ap);
			va_end(ap);
		}
		#endif

		#ifdef MYSDL_DEBUG_ERROR
		void error(int group,const char *fmt,...)
		{
			va_list ap;

			va_start(ap,fmt);
			vmessagef(dlError,group,fmt,&ap);
			va_end(ap);
		}
		#endif

		#ifdef MYSDL_DEBUG_WARNING
		void warning(int group,const char *fmt,...)
		{
			va_list ap;

			va_start(ap,fmt);
			vmessagef(dlWarning,group,fmt,&ap);
			va_end(ap);
		}
		#endif

		#ifdef MYSDL_DEBUG_INFO
		void info(int group,const char *fmt,...)
		{
			va_list ap;

			va_start(ap,fmt);
			vmessagef(dlInfo,group,fmt,&ap);
			va_end(ap);
		}
		#endif

		#ifdef MYSDL_DEBUG_DEBUG
		void debug(int group,const char *fmt,...)
		{
			va_list ap;

			va_start(ap,fmt);
			vmessagef(dlDebug,group,fmt,&ap);
			va_end(ap);
		}
		#endif

		#ifdef MYSDL_DEBUG_STACK

		/**
		* Adds a debug message to the debug message strack.
		*
		* @param msg Debug message.
		*/
		/*void addMessage(const std::string msg)
		{
			messages.push_front(msg);
	}*/

		/**
		* Removes the newest debug message from the stack.
		*/
		/*void removeMessage()
		{
			if(!messages.empty())
			{
				previousMessage=messages.front();
				messages.pop_front();
			}
			else
			{
				critical(0,"Stack trace corrupted. Previous message was: \"%s\".",previousMessage.c_str());
			}
	}*/

		/**
		* Reads the newest debug message from the stack.
		*
		* @returns Newest debug message.
		*/
		/*const std::string getLatestMessage()
		{
			if(messages.empty())
			{
				return("");
			}
			else
			{
				messages.front();
				return(messages.front());
			}
	}*/

		/**
		* Reads and removes the newest debug message from the
		* stack.
		*
		* @param s Will contain the debug message.
		*
		* @returns <code>true</code> if the stack contained a
		*          debug message. If <code>false</code> is
		*          returned, s was not changed.
		*/
		/*bool popLatestMessage(std::string &s)
		{
			if(messages.empty())
			{
				return(false);
			}
			else
			{
				s=getLatestMessage();
				removeMessage();
				return(true);
			}
	}*/

		#endif

		#ifdef MYSDL_DEBUG_PROFILE

		class ProfileData
		{
			public:
				ProfileData *parent;

				int snapshotCount;
				GLfloat beginTicks;
				GLfloat endTicks;
				GLfloat minTicks;
				GLfloat maxTicks;
				GLfloat avgTicks;
				GLfloat lastTicks;

				std::map<std::string,ProfileData> subSections;

				ProfileData();
				~ProfileData();

				void clear();
				void update();
				ProfileData *add(const std::string name);
				void print(const std::string &name);
				void clearStats();
				void clearCurrent();
		};

		#define MAX_PROFILING_SNAPSHOTS 1000

		bool profiling=false;
		ProfileData profileRoot=ProfileData();
		ProfileData *profileCurrent=&profileRoot;

		ProfileData::ProfileData() : parent(0),snapshotCount(0),beginTicks(-1),endTicks(-1),minTicks(-1),
																 maxTicks(-1),avgTicks(-1),lastTicks(-1),
																 subSections()
		{
		}

		ProfileData::~ProfileData()
		{
			//addMessage("ProfileData::~ ProfileData()");
			clear();
			//removeMessage();
		}

		void ProfileData::clear()
		{
			/** \todo ?? */
		}

		void ProfileData::update()
		{
			//addMessage("ProfileData::update()");

			if(snapshotCount<MAX_PROFILING_SNAPSHOTS)
			{
				snapshotCount++;
			}

			lastTicks=endTicks-beginTicks;

			if((lastTicks<minTicks)||(minTicks<0))
			{
				minTicks=lastTicks;
			}

			if((lastTicks>maxTicks)||(maxTicks<0))
			{
				maxTicks=lastTicks;
			}

			avgTicks=(avgTicks*(GLfloat)(snapshotCount-1)+lastTicks)/(GLfloat)snapshotCount;

			//removeMessage();
		}

		ProfileData *ProfileData::add(const std::string name)
		{
			//addMessage("ProfileData::add()");

			ProfileData &data=subSections[name];
			data.parent=this;

			//removeMessage();

			return(&data);
		}

		void ProfileData::print(const std::string &name)
		{
			//addMessage("ProfileData::print()");

			GLfloat percent;

			if(lastTicks>=0)
			{
				if(parent==0)
				{
					percent=100;
				}
				else
				{
					percent=(lastTicks/parent->lastTicks)*100;
				}

				printf("[profiling] %s (current: %dms/%.2f%%; min: %dms; max: %dms; avg: %dms)\n",
							name.c_str(),(int)lastTicks,percent,(int)minTicks,(int)maxTicks,(int)avgTicks);

				std::string space="  ";
				int i=0;

				while(name[i]==' ')
				{
					space+=' ';
					i++;
				}

				for(std::map<std::string,ProfileData>::iterator itr=subSections.begin(); itr!=subSections.end(); itr++)
				{
					std::string subName=space+itr->first;
					itr->second.print(subName);
				}
			}

			//removeMessage();
		}

		void ProfileData::clearStats()
		{
			//addMessage("ProfileData::clearStats()");

			lastTicks=minTicks=maxTicks=avgTicks=-1;

			for(std::map<std::string,ProfileData>::iterator itr=subSections.begin(); itr!=subSections.end(); itr++)
			{
				itr->second.clearStats();
			}

			//removeMessage();
		}

		void ProfileData::clearCurrent()
		{
			//addMessage("ProfileData::clearCurrent()");

			lastTicks=beginTicks=endTicks=-1;

			for(std::map<std::string,ProfileData>::iterator itr=subSections.begin(); itr!=subSections.end(); itr++)
			{
				itr->second.clearCurrent();
			}

			//removeMessage();
		}

		void beginProfileSection(const std::string name)
		{
			//addMessage("beginProfileSection()");

			if(profiling)
			{
				profileCurrent=profileCurrent->add(name);
				profileCurrent->beginTicks=SDL_GetTicks();
			}

			//removeMessage();
		}

		void endProfileSection()
		{
			//addMessage("endProfileSection()");

			if(profiling)
			{
				profileCurrent->endTicks=SDL_GetTicks();
				profileCurrent->update();
				profileCurrent=profileCurrent->parent;
			}

			//removeMessage();
		}

		void beginProfileLoop()
		{
			//addMessage("beginProfileLoop()");

			profiling=true;
			profileRoot.clearCurrent();
			profileRoot.beginTicks+=SDL_GetTicks();
			profileCurrent=&profileRoot;

			//removeMessage();
		}

		void endProfileLoop(bool incomplete)
		{
			//addMessage("endProfileLoop()");

			profileRoot.endTicks+=SDL_GetTicks();
			profileRoot.update();

			if((!incomplete)&&(profileCurrent!=&profileRoot))
			{
				printf("[profiling] (warning) Profile recursion incomplete.\n");
				fflush(stdout);
			}

			profileCurrent=&profileRoot;
			profiling=false;

			//removeMessage();
		}

		void printProfileStats()
		{
			//addMessage("printProfileStats()");

			printf("[profiling]\n[profiling] Printing profiling information:\n");
			profileRoot.print(std::string("root"));
			fflush(stdout);

			//removeMessage();
		}

		void clearProfileStats()
		{
			//addMessage("clearProfileStats()");

			if(profiling)
			{
				profileRoot.clearStats();
			}

			//removeMessage();
		}

		#endif

		void init()
		{
		#ifdef MYSDL_DEBUG_PROFILE
			profiling=false;
			profileCurrent=&profileRoot;
		#endif
		}
	};
};
