/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <iostream>
#include <vector>

#include "SDL/SDL.h"

#include "../debug.h"

#include "bitmasks.h"

namespace mysdl
{
	namespace gfx
	{
		/** Returns the minimum of two values. TODO: move to math.h */
		#define min(a,b) ((a) < (b) ? (a) : (b))

		/** Returns the maximum of two values. TODO: move to math.h */
		#define max(a,b) ((a) > (b) ? (a) : (b))

		// ================================================================
		//  Common functions
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		bool checkCollisionBitmask(Bitmask *a,Bitmask *b,Sint32 xOffset,Sint32 yOffset)
		{
			//debug::addMessage("Bitmask::checkCollisionBitmask()");

			if(xOffset<0)
			{
				Bitmask *c=a;
				a=b;
				b=c;

				xOffset=-xOffset;
				yOffset=-yOffset;
			}

			Sint32 wordOffset=xOffset/BITWIDTH;
			Sint32 bitOffset =xOffset%BITWIDTH;

			Sint32 atomsPerLineA=(a->w+MSB)/BITWIDTH;
			Sint32 atomsPerLineB=(b->w+MSB)/BITWIDTH;

			atom *line1,*line2;
			atom atom11,atom12,atom2;

			for(Sint32 y=max(0,yOffset); y<min(a->h,b->h+yOffset); y++)
			{
				line1=&(a->bits[y*atomsPerLineA]);
				line2=&(b->bits[(y-yOffset)*atomsPerLineB]);

				Sint32 to=min((b->w+MSB)/BITWIDTH,(a->w+MSB-bitOffset)/BITWIDTH-wordOffset);

				for(Sint32 x=0; x<to; x++)
				{
					//debug::addMessage("BitCore");

					atom11=line1[x+wordOffset]<<bitOffset;
					atom12=((bitOffset)&&(x<to-1) ? line1[x+wordOffset+1]>>(BITWIDTH-bitOffset) : 0);
					atom2 =line2[x];

					//debug::removeMessage();

					if((atom11&atom2)||(atom12&atom2))
					{
						//debug::removeMessage();
						return(true);
					}
				}
			}

			//debug::removeMessage();
			return(false);
		}

		/*
		 * See header file for documentation.
		 */
		void printCollisionBitmask(Bitmask *a,Bitmask *b,Sint32 xOffset,Sint32 yOffset)
		{
			Uint32 overlapWidth,overlapHeight;
			Uint32 ax1,ax2,bx1,bx2;
			Uint32 ay1,ay2,by1,by2;

			if(xOffset>=0)
			{
				overlapWidth=min((a->w-xOffset),(b->w));

				ax1=xOffset;
				ax2=min((a->w),(xOffset+b->w))-1;

				bx1=0;
				bx2=min((b->w),(a->w-xOffset))-1;
			}
			else
			{
				overlapWidth=min((b->w+xOffset),(a->w));

				ax1=0;
				ax2=min((a->w),(b->w+xOffset))-1;

				bx1=-xOffset;
				bx2=min((b->w),(-xOffset+a->w))-1;
			}

			if(yOffset>=0)
			{
				overlapHeight=min((a->h-yOffset),(b->h));

				ay1=yOffset;
				ay2=min((a->h),(yOffset+b->h))-1;

				by1=0;
				by2=min((b->h),(a->h-yOffset))-1;
			}
			else
			{
				overlapHeight=min((b->h+yOffset),(a->h));

				ay1=0;
				ay2=min((a->h),(b->h+yOffset))-1;

				by1=-yOffset;
				by2=min((b->h),(-yOffset+a->h))-1;
			}

			for(Uint32 y=0; y<overlapHeight; y++)
			{
				Uint32 ay=ay1+y;
				Uint32 by=by1+y;

				Uint32 iay=ay*((a->w+MSB)/BITWIDTH);
				Uint32 iby=by*((b->w+MSB)/BITWIDTH);

				for(Uint32 x=0; x<overlapWidth; x++)
				{
					Uint32 ax=ax1+x;
					Uint32 bx=bx1+x;

					Uint32 ia=iay+(ax/BITWIDTH);
					Uint32 ib=iby+(bx/BITWIDTH);

					Uint32 bita=ax%BITWIDTH;
					Uint32 bitb=bx%BITWIDTH;

					Uint8 bit=0;

					if(a->bits[ia]&(1<<(MSB-bita)))
					{
						bit+=1;
					}

					if(b->bits[ib]&(1<<(MSB-bitb)))
					{
						bit+=2;
					}

					std::cout << (bit==0 ? "." : (bit==3 ? "X" : (bit==1 ? "1" : "2")));
				}

				std::cout << "\n";
			}
		}

		// ================================================================
		//  Bitmask
		// ================================================================

		/** Size of the bounding box cache. TODO: ? */
		int cacheSize=0;

		/** Cache for calculating the bitmask bounding boxes. */
		int *upper=0,*lower=0;

		/*
		 * See header file for documentation.
		 */
		Bitmask::Bitmask(Sint32 w,Sint32 h) :
			w(w),h(h),bits(0),size(0),
			outerBoundingBox(),innerBoundingBox()
		{
			//debug::addMessage("Bitmask::Bitmask()");

			Uint32 atomsPerLine=(w+MSB)/BITWIDTH;
			size=atomsPerLine*h;
			bits=new atom[size];

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		Bitmask::Bitmask(SDL_Surface *surface,GLfloat threshold,bool keepCache) :
			w(surface->w),h(surface->h),bits(0),size(0),
			outerBoundingBox(),innerBoundingBox()
		{
			//debug::addMessage("Bitmask::Bitmask(SDL_Surface *,GLfloat)");

			int minX=w;
			int maxX=0;
			int minY=h;
			int maxY=0;

			Uint32 atomsPerLine=(w+MSB)/BITWIDTH;
			size=atomsPerLine*h;

			bits=new atom[size];

			clear();

			// --- Fill new bitmask ---
			int iInt=0,iBit=0;

			if(SDL_MUSTLOCK(surface))
			{
				if(SDL_LockSurface(surface))
				{
					//debug::removeMessage();
					return; // \todo throw exception
				}
			}

			if(cacheSize!=w)
			{
				clearBitmaskCache();
				cacheSize=w;

				upper=new int[w];
				lower=new int[w];
			}

			for(int x=0; x<w; x++)
			{
				upper[x]=-1;
			}

			for(int y=0; y<h; y++)
			{
				for(int x=0; x<w; x++)
				{
					bool hit=false;

					switch(surface->format->BitsPerPixel)
					{
						case 24:
						{
							break;
							Uint32 pixel=*((Uint32 *)surface->pixels+(y*surface->pitch/3+x));
							pixel>>=8;

							Uint32 alpha=(pixel)&(surface->format->Amask);
							alpha>>=surface->format->Ashift;
							alpha<<=surface->format->Aloss;

							GLfloat afloat=(GLfloat)alpha;
							afloat/=256;

							if(afloat>threshold)
							{
								hit=true;
								bits[iInt]|=(1<<(23-iBit));
							}

							iBit++;

							if(iBit>=24)
							{
								iInt++;
								iBit-=24;
							}

							break;
						}

						case 32:
						{
							Uint32 pixel=*((Uint32 *)surface->pixels+(y*surface->pitch/4+x));

							Uint32 alpha=(pixel)&(surface->format->Amask);
							alpha>>=surface->format->Ashift;
							alpha<<=surface->format->Aloss;

							GLfloat afloat=(GLfloat)alpha;
							afloat/=256;

							if(afloat>threshold)
							{
								hit=true;
								bits[iInt]|=(1<<(31-iBit));
							}

							iBit++;

							if(iBit>=32)
							{
								iInt++;
								iBit-=32;
							}

							break;
						}

						//TODO: Throw exception.
						default: printf("[bitmask] Can't handle %d bits per pixel.\n",(int)(surface->format->BitsPerPixel));
					}

					if(hit)
					{
						// --- Outer bounding box ---

						if(x>maxX)
						{
							maxX=x;
						}

						if(x<minX)
						{
							minX=x;
						}

						if(y>maxY)
						{
							maxY=y;
						}

						if(y<minY)
						{
							minY=y;
						}

						// --- Inner bounding box ---
						lower[x]=y;

						if(upper[x]==-1)
						{
							upper[x]=lower[x];
						}

						int y2=upper[x];

						for(int x2=x; x2>=0; x2--)
						{
							if(lower[x2]==-1)
							{
								break;
							}

							if(upper[x2]>y2)
							{
								y2=upper[x2];
							}

							RectInt r(x2,y2,x-x2+1,y-y2+1);

							if(r>innerBoundingBox)
							{
								innerBoundingBox=r;
							}
						}
					}
					else
					{
						upper[x]=lower[x]=-1;
					}
				}

				if(iBit!=0)
				{
					iBit=0;
					iInt++;
				}
			}

			if(SDL_MUSTLOCK(surface))
			{
				SDL_UnlockSurface(surface);
			}

			if(!keepCache)
			{
				clearBitmaskCache();
			}

			outerBoundingBox.pos.x=minX;
			outerBoundingBox.pos.y=minY;
			outerBoundingBox.dimensions.x=maxX-minX+1;
			outerBoundingBox.dimensions.y=maxY-minY+1;

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		Bitmask::~Bitmask()
		{
			delete[] bits;
		}

		/*
		 * See header file for documentation.
		 */
		void Bitmask::clear()
		{
			//debug::addMessage("Bitmask::clear()");

			for(unsigned int i=0; i<size; i++)
			{
				bits[i]=0;
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Bitmask::print()
		{
			atom *a=bits;

			printf("%d x %d:\n",w,h);

			RectInt outerBoundingBox=*(getOuterBoundingBox());
			RectInt innerBoundingBox=*(getInnerBoundingBox());

			for(Sint32 y=0; y<h; y++)
			{
				for(Sint32 x=0; x<(w+BITWIDTH-1)/BITWIDTH; x++)
				{
					atom b=*a;
					a++;

					for(Uint8 i=0; i<BITWIDTH; i++)
					{
						int xpos=x*BITWIDTH+i;

						if((xpos<outerBoundingBox.pos.x)||(xpos>=outerBoundingBox.pos.x+outerBoundingBox.dimensions.x)||
							(y<outerBoundingBox.pos.y)||(y>=outerBoundingBox.pos.y+outerBoundingBox.dimensions.y))
						{
							std::cout << "0";
						}
						else if((xpos>=innerBoundingBox.pos.x)&&(xpos<innerBoundingBox.pos.x+innerBoundingBox.dimensions.x)&&
										(y>=innerBoundingBox.pos.y)&&(y<innerBoundingBox.pos.y+innerBoundingBox.dimensions.y))
						{
							std::cout << "X";
						}
						else if(b & 0x80000000)
						{
							std::cout << "#";
						}
						else
						{
							std::cout << ".";
						}

						b<<=1;
					}
				}

				std::cout << "|\n";
			}
		}

		/*
		 * See header file for documentation.
		 */
		bool Bitmask::at(Sint32 x,Sint32 y)
		{
			//debug::addMessage("Bitmask::at()");

			Uint32 atomsPerLine=(w+MSB)/BITWIDTH;

			Sint32 offset=y*atomsPerLine+x/BITWIDTH;
			Sint32 shift =x%BITWIDTH;

			bool b=(bits[offset]<<shift) & 0x80000000;

			//debug::removeMessage();

			return(b);
		}

		/*
		 * See header file for documentation.
		 */
		const RectInt *Bitmask::getOuterBoundingBox()
		{
			return(&outerBoundingBox);
		}

		/*
		 * See header file for documentation.
		 */
		const RectInt *Bitmask::getInnerBoundingBox()
		{
			return(&innerBoundingBox);
		}

		/*
		 * See header file for documentation.
		 */
		void clearBitmaskCache()
		{
			//debug::addMessage("clearBitmaskCache()");

			if(upper)
			{
				delete[] upper;
				upper=0;
			}

			if(lower)
			{
				delete[] lower;
				lower=0;
			}

			cacheSize=0;

			//debug::removeMessage();
		}
	};
};
