/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <assert.h>
#include <iostream>
#include <map>
#include <stdarg.h>
#include <string>

#include "fonts.h"

namespace mysdl
{
	/*
	 * See header file for documentation.
	 */
	namespace fonts
	{
		// ================================================================
		//  FontResourceBankGenerator
		// ================================================================

		/**
		 * A resource bank generator for font banks.
		 */
		class FontResourceBankGenerator : public ResourceBankGenerator
		{
			public:
				BasicResourceBank *generateBank(const std::string &name) const;
		};

		/**
		 * Generates a font bank if the bank name is "fonts".
		 *
		 * @param name Name for the bank that should be generated.
		 *
		 * @return A font bank, or 0.
		 */
		BasicResourceBank *FontResourceBankGenerator::generateBank(const std::string &name) const
		{
			if(name=="fonts")
			{
				return(new fonts::FontBank());
			}
			else
			{
				return(0);
			}
		}

		/** A generator for font banks (automatically created &
		 *  registered. */
		FontResourceBankGenerator resourceBankGenerator;

		// ================================================================
		//  XML constants
		// ================================================================

		/** XML element and attribute names. */
		const char *const XML_FONTS_FONTS="fonts";
		const char *const XML_FONTS_BITMAPFONT="bitmapfont";
		const char *const XML_FONTS_DIR="dir";

		const char *const XML_FILE_BITMAPFONT="bitmapfont.xml";

		const char *const XML_FONTS_BITMAPFONT_BITMAPFONT="bitmapfont";
		const char *const XML_FONTS_BITMAPFONT_NAME="name";
		const char *const XML_FONTS_BITMAPFONT_SIZE="size";
		const char *const XML_FONTS_BITMAPFONT_SPACING="spacing";
		const char *const XML_FONTS_BITMAPFONT_YHOTSPOT="yHotSpot";
		const char *const XML_FONTS_BITMAPFONT_HPADDING="hPadding";
		const char *const XML_FONTS_BITMAPFONT_VPADDING="vPadding";
		const char *const XML_FONTS_BITMAPFONT_IMAGE="image";
		const char *const XML_FONTS_BITMAPFONT_FILE="file";
		const char *const XML_FONTS_BITMAPFONT_LINE="line";
		const char *const XML_FONTS_BITMAPFONT_CHAR="char";
		const char *const XML_FONTS_BITMAPFONT_INDEX="index";
		const char *const XML_FONTS_BITMAPFONT_WIDTH="width";
		const char *const XML_FONTS_BITMAPFONT_HEIGHT="height";

		// ================================================================
		//  FontDescriptor
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		FontDescriptor::FontDescriptor() :
			family(),size(),style()
		{
		};

		/*
		 * See header file for documentation.
		 */
		FontDescriptor::FontDescriptor(std::string family,unsigned int size,FontStyle style) :
			family(family),size(size),style(style)
		{
		};

		/*
		 * See header file for documentation.
		 */
		bool FontDescriptor::operator<(const FontDescriptor &d) const
		{
			if(family!=d.family)
			{
				return(family<d.family);
			}
			else if(size!=d.size)
			{
				return(size<d.size);
			}
			else
			{
				return(style<d.style);
			}
		}

		// ================================================================
		//  Font
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Font::Font() : NamedResource<FontBank,FontDescriptor>()
		{
			yHotSpot=0;
		}

		/*
		 * See header file for documentation.
		 */
		Font::~Font()
		{
		}

		/*
		 * See header file for documentation.
		 */
		void Font::refresh()
		{
			release();
			load();
		}

		/*
		 * See header file for documentation.
		 */
		void Font::printf(JustificationHorizontal j,const char *fmt,...)
		{
			va_list ap;
			char s[1024];

			va_start(ap,fmt);
			vsprintf(s,fmt,ap);
			va_end(ap);

			print(j,s);
		}

		/*
		 * See header file for documentation.
		 */
		void Font::print(const mysdl::Rect &rect,JustificationHorizontal jh,JustificationVertical jv,const char *s)
		{
			GLfloat x,y;

			switch(jh)
			{
				case jhLeft:
					x=rect.pos.x;
					break;

				case jhRight:
					x=rect.pos.x+rect.dimensions.x-getWidth(s);
					break;

				case jhCenter:
					x=rect.pos.x+rect.dimensions.x/2-getWidth(s)/2;
					break;

				default:
					assert(0);
			}

			switch(jv)
			{
				case jvTop:
					y=rect.pos.y+rect.dimensions.y-getHeight();
					break;

				case jvBottom:
					y=0;
					break;

				case jvCenter:
					y=rect.pos.y+rect.dimensions.y/2-getHeight()/2;
					break;

				default:
					assert(0);
			}

			glTranslatef(x,y,0);
			print(s);
			glTranslatef(-x,-y,0);
		}

		/*
		 * See header file for documentation.
		 */
		void Font::printf(const mysdl::Rect &rect,JustificationHorizontal jh,JustificationVertical jv,const char *fmt,...)
		{
			va_list ap;
			char s[1024];

			va_start(ap,fmt);
			vsprintf(s,fmt,ap);
			va_end(ap);

			print(rect,jh,jv,s);
		}

		/*
		 * See header file for documentation.
		 */
		void Font::printf(const char *fmt,...)
		{
			va_list ap;
			char s[1024];

			va_start(ap,fmt);
			vsprintf(s,fmt,ap);
			va_end(ap);

			print(s);
		}

		/*
		 * See header file for documentation.
		 */
		void Font::print(JustificationHorizontal j,const char *s)
		{
			switch(j)
			{
				case jhLeft:
					print(s);
					break;

				case jhRight:
					glTranslatef(-getWidth(s),0,0);
					print(s);
					break;

				case jhCenter:
					glTranslatef(-getWidth(s)>>1,0,0);
					print(s);
					break;

				default:
					assert(0);
			}
		}

		// ================================================================
		//  BitmapFont
		// ================================================================

		/**
		 * Helper class for loading a BitmapFont, representing
		 * the properties of one single character.
		 */
		class BitmapFontCharacterData
		{
			public:
				/** Image containing the character. */
				mysdl::gfx::SingleTextureImage *image;

				/** Position of the character on the image. */
				Uint16 x,y;

				/** Width and height of the character on the image. */
				Uint8 width,height;

				/** Vertical hotspot for this character. */
				Sint8 yHotSpot;

				/**
				 * Creates a new BitmapFontCharacterData object.
				 *
				 * @param image Image containing the character.
				 */
				BitmapFontCharacterData(mysdl::gfx::SingleTextureImage *image=0);
		};

		/*
		 * See header file for documentation.
		 */
		BitmapFontCharacterData::BitmapFontCharacterData(mysdl::gfx::SingleTextureImage *image) : image(0),x(0),y(0),width(0),height(0)
		{
			this->image=image;
		}

		/*
		 * See header file for documentation.
		 */
		BitmapFont::BitmapFont(const std::string dir) : Font(),
			dir(dir),firstChar(0),charCount(0),charHeight(0),spacing(0),listBase(0),widths(0),onlyImage(0),images(0)
		{
			if(this->dir[this->dir.size()-1]!='/')
			{
				this->dir+='/';
			}

			refresh(); // TODO: why??
		}

		/*
		 * See header file for documentation.
		 */
		void BitmapFont::load()
		{
			mysdl::debug::debug(dbgFonts,"Loading font %p from dir %s...",this,dir.c_str());

			xml::Node *root=xml::loadXML((dir+XML_FILE_BITMAPFONT).c_str());

			if((root)&&(root->name==XML_FONTS_BITMAPFONT_BITMAPFONT))
			{
				getFromNode(root,dir);
			}

			if(root)
			{
				delete root;
				root=0;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void BitmapFont::release()
		{
			if(widths)
			{
				delete[] widths;
				widths=0;
			}

			if(listBase)
			{
				glDeleteLists(listBase,charCount);
				listBase=0;
			}

			if(images)
			{
				delete images;
				images=0;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void BitmapFont::getFromNode(const xml::Node *root,const std::string dir)
		{
			//debug::addMessage("Font::getFromNode()");

			std::string dirCopy=dir;

			if(dirCopy[dirCopy.size()-1]!='/')
			{
				dirCopy+='/';
			}

			onlyImage=0;
			//images=0; // \todo ??
			images=new mysdl::gfx::ImageBank();

			const std::string *name=root->getAttribute(XML_FONTS_BITMAPFONT_NAME);

			if(!name)
			{
				debug::warning(dbgFonts,"No name specified for font. Using directory name.");
				name=&dir;
			}

			FontBank *bank=dynamic_cast<FontBank *>(owner);

			if(bank)
			{
				(*bank)-=this;
			}

			this->name.family=*name;

			this->name.size=root->getAttributeLong(XML_FONTS_BITMAPFONT_SIZE,this->name.size);

			if(bank)
			{
				(*bank)+=this;
			}

			std::map<char,BitmapFontCharacterData> dataMap;

			charHeight=root->getAttributeLong(XML_FONTS_BITMAPFONT_HEIGHT,charHeight);
			spacing   =root->getAttributeLong(XML_FONTS_BITMAPFONT_SPACING,spacing);
			yHotSpot  =root->getAttributeLong(XML_FONTS_BITMAPFONT_YHOTSPOT,yHotSpot);

			long width=root->getAttributeLong(XML_FONTS_BITMAPFONT_WIDTH,0);
			long hPadding=root->getAttributeLong(XML_FONTS_BITMAPFONT_HPADDING,0);
			long vPadding=root->getAttributeLong(XML_FONTS_BITMAPFONT_VPADDING,0);

			mysdl::gfx::SingleTextureImage *image=0;
			Uint8 min=255,max=0;
			Uint8 line=0;
			Uint16 x=0,y=0;

			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				const xml::Node *node=*itr;

				if(node->name==XML_FONTS_BITMAPFONT_CHAR)
				{
					const std::string *charString=node->getAttribute(XML_FONTS_BITMAPFONT_CHAR);

					std::string dirCopy=dir;

					if(dirCopy[dirCopy.size()-1]!='/')
					{
						dirCopy+='/';
					}

					char charIndex=' ';
					BitmapFontCharacterData data(0);

					if(!charString)
					{
						debug::error(dbgFonts,"No character index specified for blank character node. Using #32.");
					}
					else if(charString->length()==1)
					{
						charIndex=(*charString)[0];
					}
					else
					{
						charIndex=node->getAttributeLong(XML_FONTS_BITMAPFONT_CHAR,0);
					}

					data.width   =node->getAttributeLong(XML_FONTS_BITMAPFONT_WIDTH,data.width);
					data.height  =node->getAttributeLong(XML_FONTS_BITMAPFONT_HEIGHT,data.height);
					data.yHotSpot=node->getAttributeLong(XML_FONTS_BITMAPFONT_YHOTSPOT,data.yHotSpot);

					dataMap[charIndex]=data;

					if(charIndex<min)
					{
						min=charIndex;
					}

					if(charIndex>max)
					{
						max=charIndex;
					}
				}
				else if(node->name==XML_FONTS_BITMAPFONT_IMAGE)
				{
					const std::string *filename=node->getAttribute(XML_FONTS_BITMAPFONT_FILE);

					if(!filename)
					{
						debug::error(dbgFonts,"No filename specified for font image. Skipping image.");
						continue;
					}

					image=new mysdl::gfx::SimpleImage(this->name.family+*filename,dirCopy+*filename);
					(*images)+=image;
				}

				for(xml::Nodes::const_iterator itr=node->children.begin(); itr!=node->children.end(); itr++)
				{
					const xml::Node *subNode=*itr;

					if(subNode->name==XML_FONTS_BITMAPFONT_LINE)
					{
						line=subNode->getAttributeLong(XML_FONTS_BITMAPFONT_INDEX,0); // TODO: Check if exists?

						x=0;
						y=(charHeight+vPadding)*(line-1);

						for(xml::Nodes::const_iterator itr=subNode->children.begin(); itr!=subNode->children.end(); itr++)
						{
							const xml::Node *subNode2=*itr;

							if(subNode2->name==XML_FONTS_BITMAPFONT_CHAR)
							{
								const std::string *charString=subNode2->getAttribute(XML_FONTS_BITMAPFONT_CHAR);

								if(!charString)
								{
									debug::error(dbgFonts,"No character index specified for character node. Skipping character.");
									continue;
								}

								char charIndex=' ';
								BitmapFontCharacterData data(image);
								data.x=x;
								data.y=y;
								data.width=width;
								data.height=charHeight;
								data.yHotSpot=yHotSpot;

								if(charString->length()==1)
								{
									charIndex=(*charString)[0];
								}
								else
								{
									charIndex=subNode2->getAttributeLong(XML_FONTS_BITMAPFONT_CHAR,0);
								}

								data.width   =subNode2->getAttributeLong(XML_FONTS_BITMAPFONT_WIDTH,data.width);
								data.height  =subNode2->getAttributeLong(XML_FONTS_BITMAPFONT_HEIGHT,data.height);
								data.yHotSpot=subNode2->getAttributeLong(XML_FONTS_BITMAPFONT_YHOTSPOT,data.yHotSpot);

								dataMap[charIndex]=data;

								if(charIndex<min)
								{
									min=charIndex;
								}

								if(charIndex>max)
								{
									max=charIndex;
								}

								x+=data.width+hPadding;
							}
						}
					}
				}
			}

			firstChar=min;
			charCount=max-min+1;

			widths=new Uint8[charCount];

			images->refresh(0);

			listBase=glGenLists(charCount);

			for(int i=firstChar; i<firstChar+charCount; i++)
			{
				BitmapFontCharacterData data=dataMap[i];
				widths[i-firstChar]=data.width;

				glNewList(listBase+i-firstChar,GL_COMPILE);

				if(data.image)
				{
					if((onlyImage)||(images->size()==1))
					{
						onlyImage=data.image;
					}
					else
					{
						glBindTexture(GL_TEXTURE_2D,data.image->getTextureName());
					}

					glBegin(GL_QUADS);
					glTexCoord2f(((GLfloat)data.x)/((GLfloat)data.image->getWidth()),((GLfloat)data.y+(GLfloat)data.height)/((GLfloat)data.image->getHeight()));
					glVertex2i(0,-data.yHotSpot);
					glTexCoord2f(((GLfloat)data.x)/((GLfloat)data.image->getWidth()),((GLfloat)data.y)/((GLfloat)data.image->getHeight()));
					glVertex2i(0,data.height-data.yHotSpot);
					glTexCoord2f(((GLfloat)data.x+(GLfloat)data.width)/((GLfloat)data.image->getWidth()),((GLfloat)data.y)/((GLfloat)data.image->getHeight()));
					glVertex2i(data.width,data.height-data.yHotSpot);
					glTexCoord2f(((GLfloat)data.x+(GLfloat)data.width)/((GLfloat)data.image->getWidth()),((GLfloat)data.y+(GLfloat)data.height)/((GLfloat)data.image->getHeight()));
					glVertex2i(data.width,-data.yHotSpot);

					glEnd();
				}

				glTranslatef(data.width+spacing,0,0);

				glEndList();
			}
		}

		/*
		 * See header file for documentation.
		 */
		BitmapFont::~BitmapFont()
		{
			release();
		}

		/*
		 * See header file for documentation.
		 */
		Uint8 BitmapFont::getHeight()
		{
			return(charHeight);
		}

		/*
		 * See header file for documentation.
		 */
		Uint16 BitmapFont::getWidth(const char *s)
		{
			Uint16 width=0;

			for(const char *c=s; *c; c++)
			{
				if((*c>=firstChar)&&(*c<firstChar+charCount))
				{
					width+=widths[*c-firstChar];
				}

				width+=spacing;
			}

			return(width);
		}

		/*
		 * See header file for documentation.
		 */
		void BitmapFont::print(const char *s)
		{
			glPushAttrib(GL_CURRENT_BIT|GL_ENABLE_BIT|GL_TEXTURE_BIT);
			glEnable(GL_TEXTURE_2D);

			if(onlyImage)
			{
				glBindTexture(GL_TEXTURE_2D,onlyImage->getTextureName());
			}

			glListBase(listBase-firstChar);
			glCallLists(strlen(s),GL_BYTE,s);

			glPopAttrib();
		}

		// ================================================================
		//  FontBank
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		FontBank::~FontBank()
		{
			mysdl::debug::debug(dbgFonts,"Deleting font bank %p.",this);
		}

		/*
		 * See header file for documentation.
		 */
		std::string FontBank::getNodeName()
		{
			return(XML_FONTS_FONTS);
		}

		/*
		 * See header file for documentation.
		 */
		void FontBank::refresh()
		{
			// We must not refresh fonts directly while iterating through the bank,
			// as a font will remove and add itself from and to the bank during
			// refresh to change its name.

			std::list<Font *> list;

			for(iterator itr=begin(); itr!=end(); itr++)
			{
				list.push_back(*itr);
			}

			for(std::list<Font *>::iterator itr=list.begin(); itr!=list.end(); itr++)
			{
				(*itr)->refresh();
			}
		}

		/*
		 * See header file for documentation.
		 */
		void FontBank::getFromNode(const xml::Node *root,const std::string &baseDir)
		{
			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				const xml::Node *node=*itr;

				if(node->name==XML_FONTS_BITMAPFONT)
				{
					BitmapFont *font=new BitmapFont(baseDir+node->getAttribute(XML_FONTS_DIR,""));

					if(font)
					{
						(*this)+=font;
					}
					else
					{
						debug::critical(dbgFonts,"Error loading bitmap font, reason unknown.");
					}
				}
			}
		}

		// ================================================================
		//  Message
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Message::Message()
		{
			font=0;
			text="";
			ttl=0;
		}

		/*
		 * See header file for documentation.
		 */
		Message::Message(std::string text,GLfloat ttl)
		{
			font=0;
			this->text=text;
			this->ttl=ttl;
		}

		/*
		 * See header file for documentation.
		 */
		Message::~Message()
		{
		}

		/*
		 * See header file for documentation.
		 */
		void Message::perform(GLfloat ticks)
		{
			ttl-=ticks;
		}

		/*
		 * See header file for documentation.
		 */
		void Message::draw(Rect rect,JustificationHorizontal jh,JustificationVertical jv)
		{
			font->print(rect,jh,jv,text.c_str());
		}

		// ================================================================
		//  BlinkMessage
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		BlinkMessage::BlinkMessage() : Message()
		{
			visible=true;
			blinkCounter=blinkDurationOn=blinkDurationOff=0;
		}

		/*
		 * See header file for documentation.
		 */
		BlinkMessage::BlinkMessage(std::string text,GLfloat ttl,GLfloat blinkDurationOn,GLfloat blinkDurationOff) : Message(text,ttl)
		{
			visible=true;
			blinkCounter=blinkDurationOn;
			this->blinkDurationOn=blinkDurationOn;
			this->blinkDurationOff=(blinkDurationOff<0 ? blinkDurationOn : blinkDurationOff);
		}

		/*
		 * See header file for documentation.
		 */
		void BlinkMessage::perform(GLfloat ticks)
		{
			Message::perform(ticks);

			blinkCounter-=ticks;

			while(blinkCounter<0)
			{
				visible=!visible;
				blinkCounter+=(visible ? blinkDurationOn : blinkDurationOff);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void BlinkMessage::draw(Rect rect,JustificationHorizontal jh,JustificationVertical jv)
		{
			if(visible)
			{
				Message::draw(rect,jh,jv);
			}
		}

		// ================================================================
		//  Messages
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Messages::Messages() : messages(),rect(),font(0),multiple(false)
		{
		}

		/*
		 * See header file for documentation.
		 */
		Messages::~Messages()
		{
			clear();
		}

		/*
		 * See header file for documentation.
		 */
		void Messages::operator+=(Message *message)
		{
			if(!multiple)
			{
				clear();
			}

			messages.push_back(message);

			if(!message->font)
			{
				message->font=font;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Messages::perform(GLfloat ticks)
		{
			for(std::list<Message *>::iterator itr=messages.begin(); itr!=messages.end(); itr++)
			{
				(*itr)->perform(ticks);
			}

			purge();
		}

		/*
		 * See header file for documentation.
		 */
		void Messages::draw()
		{
			for(std::list<Message *>::iterator itr=messages.begin(); itr!=messages.end(); itr++)
			{
				(*itr)->draw(rect,jhCenter,jvCenter);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Messages::clear()
		{
			for(std::list<Message *>::iterator itr=messages.begin(); itr!=messages.end(); itr++)
			{
				delete *itr;
			}

			messages.clear();
		}

		/*
		 * See header file for documentation.
		 */
		void Messages::purge()
		{
			std::list<Message *> toDelete;

			for(std::list<Message *>::iterator itr=messages.begin(); itr!=messages.end(); itr++)
			{
				if((*itr)->ttl<=0)
				{
					toDelete.push_back(*itr);
				}
			}

			for(std::list<Message *>::iterator itr=toDelete.begin(); itr!=toDelete.end(); itr++)
			{
				delete *itr;
				messages.remove(*itr);
			}
		}
	};
};
