/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>

#include <string>
#include <iostream>
#include <fstream>
#include <list>
#include <stack>

#include "SDL/SDL.h"
#include "SDL/SDL_image.h"

#include "../debug.h"
#include "../mysdl.h"
#include "../tools.h"
#include "../xml.h"

#include "images.h"

using namespace std;

namespace mysdl
{
	namespace gfx
	{
		// ================================================================
		//  ImageResourceBankGenerator
		// ================================================================

		/**
		 * A resource bank generator for image banks.
		 */
		class ImageResourceBankGenerator : public ResourceBankGenerator
		{
			public:
				BasicResourceBank *generateBank(const std::string &name) const;
		};

		/**
		 * Generates an image bank if the bank name is "images".
		 *
		 * @param name Name for the bank that should be generated.
		 *
		 * @return An image bank, or 0.
		 */
		BasicResourceBank *ImageResourceBankGenerator::generateBank(const std::string &name) const
		{
			if(name=="images")
			{
				return(new ImageBank());
			}
			else
			{
				return(0);
			}
		}

		/** A generator for image banks (automatically created &
		 *  registered. */
		ImageResourceBankGenerator imageResourceBankGenerator;

		// ================================================================
		//  XML constants
		// ================================================================

		/** XML element and attribute names. */
		const char *const XML_IMAGES_IMAGES           ="images";
		const char *const XML_IMAGES_IMAGE            ="image";
		const char *const XML_IMAGES_NAME             ="name";
		const char *const XML_IMAGES_FILENAME         ="filename";
		const char *const XML_IMAGES_HOTX             ="hotX";
		const char *const XML_IMAGES_HOTY             ="hotY";
		const char *const XML_IMAGES_BITMASKTHRESHOLD ="bitmaskThreshold";
		const char *const XML_IMAGES_CLAMP            ="clamp";
		const char *const XML_IMAGES_FILTER           ="filter";

		const char *const XML_IMAGES_LOOP             ="loop";
		const char *const XML_IMAGES_VAR              ="var";
		const char *const XML_IMAGES_FROM             ="from";
		const char *const XML_IMAGES_TO               ="to";
		const char *const XML_IMAGES_STEP             ="step";
		const char *const XML_IMAGES_DIGITS           ="digits";

		// ================================================================
		//  Image
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Image::Image() : width(0),height(0),hotspot(),bitmask(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		Image::~Image()
		{
		}

		/*
		 * See header file for documentation.
		 */
		void Image::perform(GLfloat ticks)
		{
		}

		/*
		 * See header file for documentation.
		 */
		Uint32 Image::getWidth() const
		{
			return(width);
		}

		/*
		 * See header file for documentation.
		 */
		Uint32 Image::getHeight() const
		{
			return(height);
		}

		/*
		 * See header file for documentation.
		 */
		Bitmask *Image::getBitmask() const
		{
			return(bitmask);
		}

		/*
		 * See header file for documentation.
		 */
		const RectInt *Image::getOuterBoundingBox()
		{
			return(&outerBoundingBox);
		}

		/*
		 * See header file for documentation.
		 */
		const RectInt *Image::getInnerBoundingBox()
		{
			return(&innerBoundingBox);
		}

		/*
		 * See header file for documentation.
		 */
		void Image::setHotspot(Vector2D &hotspot)
		{
			this->hotspot=hotspot;
		}

		/*
		 * See header file for documentation.
		 */
		const Vector2D &Image::getHotspot() const
		{
			return(hotspot);
		}

		// ================================================================
		//  ImageResource
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		ImageResource::ImageResource() :
			NamedResource<ImageBank>(),Image(),filename(),bitmaskThreshold(-1),clamp(false),filter(ifLinear)
		{
		}

		/*
		 * See header file for documentation.
		 */
		ImageResource::ImageResource(const std::string &name,const std::string &filename) :
			NamedResource<ImageBank>(name),Image(),filename(filename),bitmaskThreshold(-1),clamp(false),filter(ifLinear)
		{
		}

		/*
		 * See header file for documentation.
		 */
		ImageResource::~ImageResource()
		{
			if(bitmask)
			{
				delete bitmask;
				bitmask=0;
			}

			release();
		}

		/*
		 * See header file for documentation.
		 */
		void ImageResource::release(bool last)
		{
		}

		/*
		 * See header file for documentation.
		 */
		ImageFilter ImageResource::getFilter() const
		{
			return(filter);
		}

		/*
		 * See header file for documentation.
		 */
		void ImageResource::setFilename(std::string filename)
		{
			this->filename=filename;
		}

		/*
		 * See header file for documentation.
		 */
		std::string ImageResource::getFilename()
		{
			return(filename);
		}

		/*
		 * See header file for documentation.
		 */
		void ImageResource::setBitmaskThreshold(GLfloat bitmaskThreshold)
		{
			this->bitmaskThreshold=bitmaskThreshold;
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat ImageResource::getBitmaskThreshold()
		{
			return(bitmaskThreshold);
		}

		/*
		 * See header file for documentation.
		 */
		void ImageResource::setClamp(bool clamp)
		{
			this->clamp=clamp;
		}

		/*
		 * See header file for documentation.
		 */
		bool ImageResource::getClamp() const
		{
			return(clamp);
		}

		/*
		 * See header file for documentation.
		 */
		void ImageResource::reload(bool last)
		{
			load(last,bitmask==0);
		}

		// ================================================================
		//  SingleTextureImage
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		SingleTextureImage::SingleTextureImage() :
			ImageResource(),textureName(0),texCoordRect(),textureWidth(0),textureHeight(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		SingleTextureImage::SingleTextureImage(const std::string &name,const std::string &filename) :
			ImageResource(name,filename),textureName(0),texCoordRect(),textureWidth(0),textureHeight(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		GLuint SingleTextureImage::getTextureName() const
		{
			return(textureName);
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat SingleTextureImage::getLeftTexCoord() const
		{
			return(texCoordRect.pos.x);
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat SingleTextureImage::getRightTexCoord() const
		{
			return(texCoordRect.pos.x+texCoordRect.dimensions.x);
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat SingleTextureImage::getUpperTexCoord() const
		{
			return(texCoordRect.pos.y);
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat SingleTextureImage::getLowerTexCoord() const
		{
			return(texCoordRect.pos.y+texCoordRect.dimensions.y);
		}

		/*
		 * See header file for documentation.
		 */
		const Rect &SingleTextureImage::getTexCoordRect() const
		{
			return(texCoordRect);
		}

		/*
		 * See header file for documentation.
		 */
		void SingleTextureImage::draw(const Vector2D &pos,BlendMode blendMode) const
		{
			glPushMatrix();
			glPushAttrib(GL_TEXTURE_2D|GL_COLOR);

			if(blendMode!=bmDefault)
			{
				setBlendMode(blendMode);
			}

			glEnable(GL_TEXTURE_2D);
			glTranslatef(pos.x-hotspot.x,pos.y+hotspot.y,0);
			glBindTexture(GL_TEXTURE_2D,textureName);

			glBegin(GL_QUADS);
			glTexCoord2f(getLeftTexCoord(),getLowerTexCoord());
			glVertex2f(0,0);

			glTexCoord2f(getLeftTexCoord(),getUpperTexCoord());
			glVertex2f(0,height);

			glTexCoord2f(getRightTexCoord(),getUpperTexCoord());
			glVertex2f(width,height);

			glTexCoord2f(getRightTexCoord(),getLowerTexCoord());
			glVertex2f(width,0);
			glEnd();

			glPopAttrib();
			glPopMatrix();
		}

		/*
		 * See header file for documentation.
		 */
		void SingleTextureImage::draw(const Rect &rect,BlendMode blendMode) const
		{
			GLfloat xMax=rect.pos.x+rect.dimensions.x;
			GLfloat yMax=rect.pos.y+rect.dimensions.y;

			glPushAttrib(GL_TEXTURE_2D|GL_COLOR);

			if(blendMode!=bmDefault)
			{
				setBlendMode(blendMode);
			}

			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D,textureName);

			if((textureWidth==width)&&(textureHeight==height))
			{
				glBegin(GL_QUADS);
				glTexCoord2f(0,0);
				glVertex2f(rect.pos.x,rect.pos.y+rect.dimensions.y);

				glTexCoord2f(0,rect.dimensions.y/(GLfloat)textureHeight); // \todo -1?
				glVertex2f(rect.pos.x,rect.pos.y);

				glTexCoord2f(rect.dimensions.x/(GLfloat)textureWidth,rect.dimensions.y/(GLfloat)textureHeight);
				glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y);

				glTexCoord2f(rect.dimensions.x/(GLfloat)textureWidth,0);
				glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y);
				glEnd();
			}
			/*
			else if(textureWidth==width)
			{
				glBegin(GL_QUADS);

				for(GLfloat y=rect.pos.y; y<rect.pos.y+rect.dimensions.y; y+=height)
				{
					glTexCoord2f(getLeftTexCoord(),getLowerTexCoord());
					glVertex2f(rect.pos.x,y);

					glTexCoord2f(getLeftTexCoord(),getUpperTexCoord());
					glVertex2f(rect.pos.x,min(y+height,yMax));

					glTexCoord2f(getRightTexCoord(),getUpperTexCoord());
					glVertex2f(min(rect.pos.x+rect.dimensions.x,xMax),min(y+height,yMax));

					glTexCoord2f(getRightTexCoord(),getLowerTexCoord());
					glVertex2f(min(rect.pos.x+rect.dimensions.x,xMax),y);
				}

				glEnd();
			}
			else if(textureHeight==height)
			{
				glBegin(GL_QUADS);

				for(GLfloat x=rect.pos.x; x<rect.pos.x+rect.dimensions.x; x+=width)
				{
					glTexCoord2f(getLeftTexCoord(),getLowerTexCoord());
					glVertex2f(x,rect.pos.y);

					glTexCoord2f(getLeftTexCoord(),getUpperTexCoord());
					glVertex2f(x,min(rect.pos.y+rect.dimensions.y,yMax));

					glTexCoord2f(getRightTexCoord(),getUpperTexCoord());
					glVertex2f(min(x+width,xMax),min(rect.pos.y+rect.dimensions.y,yMax));

					glTexCoord2f(getRightTexCoord(),getLowerTexCoord());
					glVertex2f(min(x+width,xMax),rect.pos.y);
				}

				glEnd();
			}
			*/

/*
			else
			{
				glBegin(GL_QUADS);

				GLfloat leftTexCoord=getLeftTexCoord();
				GLfloat lowerTexCoord=getLowerTexCoord();

				for(GLfloat y=rect.pos.y; y<yMax; y+=height)
				{
					GLfloat upperTexCoord=getUpperTexCoord();
					GLfloat top=y+height;

					if(top>yMax)
					{
						top=yMax;
						lowerTexCoord=lowerTexCoord*((yMax-y)/(GLfloat)height);
					}

					for(GLfloat x=rect.pos.x; x<xMax; x+=width)
					{
						GLfloat rightTexCoord=getRightTexCoord();

						GLfloat right=x+width;

						if(right>xMax)
						{
							right=xMax;
							rightTexCoord=rightTexCoord*((xMax-x)/(GLfloat)width);
						}

						glTexCoord2f(leftTexCoord,lowerTexCoord);
						glVertex2f(x,y);

						glTexCoord2f(leftTexCoord,upperTexCoord);//getUpperTexCoord());
						glVertex2f(x,top);

						glTexCoord2f(rightTexCoord,upperTexCoord);//getUpperTexCoord());
						glVertex2f(right,top);

						glTexCoord2f(rightTexCoord,lowerTexCoord);
						glVertex2f(right,y);
					}
				}

				glEnd();
			}
*/
			else
			{
				glBegin(GL_QUADS);

				GLfloat leftTexCoord=getLeftTexCoord();
				GLfloat lowerTexCoord=getLowerTexCoord();

				for(GLfloat y=yMax; y>rect.pos.y; y-=height)
				{
					GLfloat upperTexCoord=getUpperTexCoord();
					GLfloat bottom=y-height;

					if(bottom<rect.pos.y)
					{
						bottom=rect.pos.y;
						lowerTexCoord=lowerTexCoord*((y-rect.pos.y)/(GLfloat)height);
					}

					for(GLfloat x=rect.pos.x; x<xMax; x+=width)
					{
						GLfloat rightTexCoord=getRightTexCoord();

						GLfloat right=x+width;

						if(right>xMax)
						{
							right=xMax;
							rightTexCoord=rightTexCoord*((xMax-x)/(GLfloat)width);
						}

						glTexCoord2f(leftTexCoord,lowerTexCoord);
						glVertex2f(x,bottom);

						glTexCoord2f(leftTexCoord,upperTexCoord);//getUpperTexCoord());
						glVertex2f(x,y);

						glTexCoord2f(rightTexCoord/*getRightTexCoord()*/,upperTexCoord);//getUpperTexCoord());
						glVertex2f(right,y);

						glTexCoord2f(rightTexCoord/*getRightTexCoord()*/,lowerTexCoord);
						glVertex2f(right,bottom);
					}
				}

				glEnd();
			}

			glPopAttrib();
		}

		/*
		 * See header file for documentation.
		 */
		void SingleTextureImage::drawStretched(const Rect &rect,BlendMode blendMode) const
		{
			glPushAttrib(GL_TEXTURE_2D|GL_COLOR);

			if(blendMode!=bmDefault)
			{
				setBlendMode(blendMode);
			}

			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D,textureName);

			glBegin(GL_QUADS);
/*			glTexCoord2f(getLeftTexCoord(),getLowerTexCoord());
			glVertex2f(rect.pos.x,rect.pos.y+rect.dimensions.y);

			glTexCoord2f(getLeftTexCoord(),getUpperTexCoord());
			glVertex2f(rect.pos.x,rect.pos.y); // \todo -1?

			glTexCoord2f(getRightTexCoord(),getUpperTexCoord());
			glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y);

			glTexCoord2f(getRightTexCoord(),getLowerTexCoord());
			glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y);*/
			glTexCoord2f(getLeftTexCoord(),getUpperTexCoord());
			glVertex2f(rect.pos.x,rect.pos.y+rect.dimensions.y);

			glTexCoord2f(getLeftTexCoord(),getLowerTexCoord());
			glVertex2f(rect.pos.x,rect.pos.y); // \todo -1?

			glTexCoord2f(getRightTexCoord(),getLowerTexCoord());
			glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y);

			glTexCoord2f(getRightTexCoord(),getUpperTexCoord());
			glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y);
			glEnd();

			glPopAttrib();
		}

		// ================================================================
		//  SimpleImage
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		SimpleImage::SimpleImage() : SingleTextureImage()
		{
		}

		/*
		 * See header file for documentation.
		 */
		SimpleImage::SimpleImage(const std::string &name,const std::string &filename) :
			SingleTextureImage(name,filename)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void SimpleImage::release(bool last)
		{
			if(textureName)
			{
				glDeleteTextures(1,&textureName);
				textureName=0;
			}

			ImageResource::release(last);
		}

		/*
		 * See header file for documentation.
		 */
		void SimpleImage::load(bool last,bool createBitmask)
		{
			release(last);

			SDL_Surface *temp=IMG_Load(filename.c_str());

			if(!temp)
			{
				throw std::string("Could not load \""+filename+"\": "+SDL_GetError()+".");
			}

			outerBoundingBox.pos.x=0;
			outerBoundingBox.pos.y=0;
			outerBoundingBox.dimensions.x=width;
			outerBoundingBox.dimensions.y=height;

			innerBoundingBox.pos.x=0;
			innerBoundingBox.pos.y=0;
			innerBoundingBox.dimensions.x=0;
			innerBoundingBox.dimensions.y=0;

			if((createBitmask)&&(bitmaskThreshold>=0))
			{
				this->createBitmask(temp,last);
			}

			width=temp->w;
			height=temp->h;

			textureWidth=0;
			textureHeight=0;

			for(unsigned int i=1,j=2; j>0; i*=2)
			{
				if((!(textureWidth))&&(i>=width))
				{
					textureWidth=i;
					j--;
				}

				if((!(textureHeight))&&(i>=height))
				{
					textureHeight=i;
					j--;
				}
			}

			if((textureWidth!=width)||(textureHeight!=height))
			{
				SDL_Surface *temp2=SDL_CreateRGBSurface(0,textureWidth,textureHeight,
																								temp->format->BitsPerPixel,
																								temp->format->Rmask,
																								temp->format->Gmask,
																								temp->format->Bmask,
																								temp->format->Amask);

				SDL_FillRect(temp2,0,SDL_MapRGBA(temp2->format,0,0,0,0));

				SDL_SetAlpha(temp,0,0);

				if(SDL_BlitSurface(temp,0,temp2,0)!=0)
				{
					printf("Blitting failed: %s.\n",SDL_GetError()); // TODO ??
				}

				SDL_FreeSurface(temp);
				temp=temp2;
			}

			texCoordRect.dimensions.x=(GLfloat)width/(GLfloat)textureWidth;
			texCoordRect.dimensions.y=(GLfloat)height/(GLfloat)textureHeight;

			if(!textureName)
			{
				glGenTextures(1,&textureName);
			}

			glBindTexture(GL_TEXTURE_2D,textureName);

			if(filter>ifLinear)
			{
				gluBuild2DMipmaps(GL_TEXTURE_2D,
													temp->format->BitsPerPixel==32 ? GL_RGBA8 : GL_RGB8,
													temp->w,temp->h,
													temp->format->BitsPerPixel==32 ? GL_RGBA : GL_RGB,
													GL_UNSIGNED_BYTE,temp->pixels);
			}
			else
			{
				glTexImage2D(GL_TEXTURE_2D,0,
										temp->format->BitsPerPixel==32 ? GL_RGBA8 : GL_RGB8,
										temp->w,temp->h,0,
										temp->format->BitsPerPixel==32 ? GL_RGBA : GL_RGB,
										GL_UNSIGNED_BYTE,temp->pixels);
			}

			if(clamp)
			{
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
			}

			SDL_UnlockSurface(temp);
			SDL_FreeSurface(temp);
		}

		/*
		 * See header file for documentation.
		 */
		void SimpleImage::setFilter(ImageFilter filter)
		{
			if(!textureName)
			{
				return;
			}

			glBindTexture(GL_TEXTURE_2D,textureName);

			if(ImageBank *bank=owner)
			{
				if(bank->filter<filter)
				{
					filter=bank->filter;
				}
			}

			if(filter==ifNone)
			{
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
			}
			else if(filter==ifLinear)
			{
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			}
			else if(filter==ifBilinear)
			{
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST_MIPMAP_LINEAR); // \todo other way round?
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			}
			else if(filter==ifTrilinear)
			{
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void SimpleImage::refresh(bool last)
		{
			if(textureName==0)
			{
				load(last,bitmask==0);
			}

			setFilter(filter);
		}

		/*
		 * See header file for documentation.
		 */
		void ImageResource::createBitmask(SDL_Surface *surface,bool last)
		{
			if(bitmaskThreshold<0)
			{
				return;
			}

			// --- Delete old bitmask ---
			if(bitmask)
			{
				delete bitmask;
				bitmask=0;
			}

			// --- Reserve memory for new bitmask ---
			bitmask=new Bitmask(surface,bitmaskThreshold,!last);

			outerBoundingBox=*(bitmask->getOuterBoundingBox());
			innerBoundingBox=*(bitmask->getInnerBoundingBox());
		}

		// ================================================================
		//  ImageBank
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		ImageBank::ImageBank() :
			NamedResourceBank<ImageResource>(),filter(ifLinear)
		{
		}

		/*
		 * See header file for documentation.
		 */
		ImageBank::~ImageBank()
		{
		}

		/*
		 * See header file for documentation.
		 */
		void ImageBank::reload()
		{
			for(iterator itr=begin(); itr!=end(); itr++)
			{
				itr->reload(itr==end());
			}
		}

		/*
		 * See header file for documentation.
		 */
		void ImageBank::refresh(RefreshingBankHandler *handler)
		{
			for(iterator itr=begin(); itr!=end(); itr++)
			{
				itr->refresh(itr==end());

				if(handler)
				{
					handler->elementRefreshed();
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void ImageBank::release()
		{
			for(iterator itr=begin(); itr!=end(); itr++)
			{
				itr->release(itr!=end());
			}
		}

		/*
		 * See header file for documentation.
		 */
		void ImageBank::getFromNode(const xml::Node *root,const std::string &baseDir)
		{
			static std::list<mysdl::xml::LoopData *> loops; // \note: Not thread safe!

			for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
			{
				const xml::Node *node=*itr;

				if(node->name==XML_IMAGES_LOOP)
				{
					const std::string *var=node->getAttribute(XML_IMAGES_VAR);

					if(!var)
					{
						debug::error(dbgImages,"No variable specified for loop. Skipping loop.");
						continue;
					}

					int min   =(int)(node->getAttributeLong(XML_IMAGES_FROM,0));
					int max   =(int)(node->getAttributeLong(XML_IMAGES_TO,0));
					int step  =(int)(node->getAttributeLong(XML_IMAGES_STEP,1));
					int digits=(int)(node->getAttributeLong(XML_IMAGES_DIGITS,0));

					if(step==0)
					{
						debug::error(dbgAnimations,"Step size 0 invalid: %s. Skipping loop.",var->c_str());
					}

					if(((step>0)&&(max<=min))||((step<0)&&(max>=min)))
					{
						debug::error(dbgAnimations,"Loop range invalid: %s [%d;%d] (step: %d). Skipping loop.",var->c_str(),min,max,step);
						continue;
					}

					mysdl::xml::LoopData *entry=new mysdl::xml::LoopData(*var,min,max,step,digits);
					loops.push_back(entry);

					do
					{
						getFromNode(node,baseDir);
					}
					while(entry->advance());

					loops.pop_back();
					delete entry;
				}
				else if(node->name==XML_IMAGES_IMAGE)
				{
					const std::string *namePtr=node->getAttribute(XML_IMAGES_NAME);
					const std::string *filenamePtr=node->getAttribute(XML_IMAGES_FILENAME);

					if(!namePtr)
					{
						debug::warning(dbgImages,"No name specified for image. Using filename.");
						namePtr=filenamePtr;
					}

					if(!filenamePtr)
					{
						debug::error(dbgImages,"No filename specified for image \"%s\". Skipping entry.",namePtr->c_str());
						continue;
					}

					std::string name=*namePtr;
					std::string filename=*filenamePtr;

					for(std::list<mysdl::xml::LoopData *>::iterator itr=loops.begin(); itr!=loops.end(); itr++)
					{
						std::string varStr("["+(*itr)->var+"]");
						std::string valStr((*itr)->getString());

						for(std::string::size_type i=name.find(varStr); i!=std::string::npos; i=name.find(varStr))
						{
							name.replace(i,varStr.size(),valStr);
						}

						for(std::string::size_type i=filename.find(varStr); i!=std::string::npos; i=filename.find(varStr))
						{
							filename.replace(i,varStr.size(),valStr);
						}
					}

					SimpleImage *image=new SimpleImage(name,baseDir+filename);

					image->hotspot.x=node->getAttributeFloat(XML_IMAGES_HOTX,image->hotspot.x);
					image->hotspot.y=node->getAttributeFloat(XML_IMAGES_HOTY,image->hotspot.y);

					image->bitmaskThreshold=node->getAttributeFloat(XML_IMAGES_BITMASKTHRESHOLD,image->bitmaskThreshold);
					image->clamp=node->getAttributeBool(XML_IMAGES_CLAMP,image->clamp);

					if(const std::string *filterName=node->getAttribute(XML_IMAGES_FILTER))
					{
						for(ImageFilter filter=ifNone; filter<=ifTrilinear; filter=(ImageFilter)(((int)(filter))+1))
						{
							if(*filterName==STR_IMAGEFILTER[filter])
							{
								image->filter=filter;
							}
						}
					}

					(*this)+=image;
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		std::string ImageBank::getNodeName()
		{
			return(XML_IMAGES_IMAGES);
		}

		/*
		 * See header file for documentation.
		 */
		int ImageBank::size() const
		{
			return(NamedResourceBank<ImageResource>::size());
		}

		// ================================================================
		//  ImageLoader
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		ImageLoader::ImageLoader(ImageBank *images,BackgroundTaskCallback *callback) :
			BackgroundTask(callback),
			images(images)
		{
			totalCount=images->size();
			itr=images->begin();
		}

		/*
		 * See header file for documentation.
		 */
		void ImageLoader::perform(Uint32 timeLimit)
		{
			while(itr!=images->end())
			{
				if(SDL_GetTicks()>=timeLimit)
				{
					break;
				}

				itr->refresh(itr==images->end());
				itr++;
				finishedCount++;
			}

			if(itr==images->end())
			{
				finish();
			}
		}
	};
};
