/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <assert.h>

#include "core.h"
#include "styles.h"
#include "xml.h"

namespace mysdl
{
	template<> mysdl::gui::StretchMode strTo<mysdl::gui::StretchMode>(std::string str)
	{
		if(str=="single")
		{
			return(mysdl::gui::smSingle);
		}
		else if(str=="repeat")
		{
			return(mysdl::gui::smRepeat);
		}
		else if(str=="stretch")
		{
			return(mysdl::gui::smStretch);
		}
		else
		{
			throw(str);
		}
	}

	namespace gui
	{
		// ================================================================
		//  FontStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		FontStyle::FontStyle() : Style(),justificationHorizontal(jhLeft),justificationVertical(jvCenter)
		{
		}

		// ================================================================
		//  SimpleFontStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		SimpleFontStyle::SimpleFontStyle() : FontStyle(),font(0),color(1.0,1.0,1.0,1.0)
		{
		}

		// ================================================================
		//  RectStyle
		// ================================================================

		RectStyle::RectStyle() : Style()
		{
		}

		// ================================================================
		//  FontStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		void FontStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("FontStyle::getFromNode()");

			Style::getFromNode(node,gui);

			justificationHorizontal=mysdl::justificationHorizontal(
															node->getAttribute(XML_GUI_JUSTIFICATION_HORIZONTAL,
															STR_JUSTIFICATION_HORIZONTAL[justificationHorizontal]).c_str());

			justificationVertical  =mysdl::justificationVertical(
															node->getAttribute(XML_GUI_JUSTIFICATION_VERTICAL,
															STR_JUSTIFICATION_VERTICAL[justificationVertical]).c_str());

			//debug::removeMessage();
		}

		// ================================================================
		//  SimpleFontStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		void SimpleFontStyle::print(const Rect &rect,const char *string)
		{
			//debug::addMessage("SimpleFontStyle::print()");

			glPushAttrib(GL_CURRENT_BIT|GL_ENABLE_BIT|GL_TEXTURE_BIT);
			glColor4f(color.r,color.g,color.b,color.a);
			font->print(rect,justificationHorizontal,justificationVertical,string);
			glPopAttrib();

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void SimpleFontStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("SimpleFontStyle::getFromNode()");

			FontStyle::getFromNode(node,gui);

			if(const std::string *color=node->getAttribute(XML_GUI_COLOR))
			{
				this->color.set((*color).c_str());
			}

			if(const std::string *family=node->getAttribute(XML_GUI_FAMILY))
			{
				unsigned int size=node->getAttributeLong(XML_GUI_SIZE,0);
				this->font=(*gui->getFonts())[fonts::FontDescriptor(*family,size)];
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  PlainRectStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		PlainRectStyle::PlainRectStyle() : RectStyle(),color(),border(0)
		{
			colorBorder[0]=colorBorder[1]=Color();
		}

		/*
		 * See header file for documentation.
		 */
		void PlainRectStyle::draw(const Rect &rect)
		{
			//debug::addMessage("PlainRectStyle::draw()");

			glPushAttrib(GL_CURRENT_BIT|GL_ENABLE_BIT|GL_TEXTURE_BIT);

			glDisable(GL_TEXTURE_2D);
			glColor4f(color.r,color.g,color.b,color.a);

			glBegin(GL_QUADS);
			glVertex2f(rect.pos.x,rect.pos.y);
			glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y);
			glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y);
			glVertex2f(rect.pos.x,rect.pos.y+rect.dimensions.y);
			glEnd();

			if(border)
			{
				glColor4f(colorBorder[0].r,colorBorder[0].g,colorBorder[0].b,colorBorder[0].a);

				glBegin(GL_QUADS);
				glVertex2f(rect.pos.x-border,rect.pos.y+rect.dimensions.y+border);
				glVertex2f(rect.pos.x+rect.dimensions.x+border,rect.pos.y+rect.dimensions.y+border);
				glVertex2f(rect.pos.x+rect.dimensions.x/*+1*/,rect.pos.y+rect.dimensions.y);
				glVertex2f(rect.pos.x-border,rect.pos.y+rect.dimensions.y);

				glVertex2f(rect.pos.x-border,rect.pos.y+rect.dimensions.y);
				glVertex2f(rect.pos.x,rect.pos.y+rect.dimensions.y);
				glVertex2f(rect.pos.x,rect.pos.y/*-1*/);
				glVertex2f(rect.pos.x-border,rect.pos.y-border);
				glEnd();

				glColor4f(colorBorder[1].r,colorBorder[1].g,colorBorder[1].b,colorBorder[1].a);

				glBegin(GL_QUADS);
				glVertex2f(rect.pos.x,rect.pos.y);
				glVertex2f(rect.pos.x+rect.dimensions.x+border,rect.pos.y);
				glVertex2f(rect.pos.x+rect.dimensions.x+border,rect.pos.y-border);
				glVertex2f(rect.pos.x-border,rect.pos.y-border);

				glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y);
				glVertex2f(rect.pos.x+rect.dimensions.x+border,rect.pos.y+rect.dimensions.y+border);
				glVertex2f(rect.pos.x+rect.dimensions.x+border,rect.pos.y);
				glVertex2f(rect.pos.x+rect.dimensions.x,rect.pos.y);
				glEnd();
			}

			glPopAttrib();

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void PlainRectStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("PlainRectStyle::getFromNode()");

			RectStyle::getFromNode(node,gui);

			border=node->getAttributeFloat(XML_GUI_BORDER,border);

			if(const std::string *color=node->getAttribute(XML_GUI_COLOR))
			{
				this->color.set((*color).c_str());
			}

			if(const std::string *color=node->getAttribute(XML_GUI_COLORBORDER1))
			{
				this->colorBorder[0].set((*color).c_str());
			}

			if(const std::string *color=node->getAttribute(XML_GUI_COLORBORDER2))
			{
				this->colorBorder[1].set((*color).c_str());
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  TexturedRectStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		TexturedRectStyle::TexturedRectStyle() : RectStyle(),color(),
			imgUpperLeft(0),imgUpper(0),imgUpperRight(0),
			imgLeft(0),imgCenter(0),imgRight(0),
			imgLowerLeft(0),imgLower(0),imgLowerRight(0),
			stretchUpper(smRepeat),stretchLower(smRepeat),stretchLeft(smRepeat),stretchRight(smRepeat),stretchCenter(smRepeat)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void TexturedRectStyle::draw(const Rect &rect)
		{
			//debug::addMessage("TexturedRectStyle::draw()");

			glPushAttrib(GL_CURRENT_BIT|GL_ENABLE_BIT|GL_TEXTURE_BIT);

			glEnable(GL_TEXTURE_2D);
			glColor4f(color.r,color.g,color.b,color.a);

			if(imgUpperLeft)
			{
				imgUpperLeft->draw(Vector2D(rect.pos.x-imgUpperLeft->getWidth(),rect.pos.y+rect.dimensions.y));
			}

			if(imgUpper)
			{
				switch(stretchUpper)
				{
					case smSingle:
						imgUpper->draw(Vector2D(rect.pos.x,rect.pos.y+rect.dimensions.y));
						break;

					case smRepeat:
						imgUpper->draw(Rect(Vector2D(rect.pos.x,rect.pos.y+rect.dimensions.y),Vector2D(rect.dimensions.x,imgUpper->getHeight())));
						break;

					case smStretch:
						imgUpper->drawStretched(Rect(Vector2D(rect.pos.x,rect.pos.y+rect.dimensions.y),Vector2D(rect.dimensions.x,imgUpper->getHeight())));
						break;
				}
			}

			if(imgUpperRight)
			{
				imgUpperRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y));
			}

			if(imgLeft)
			{
				switch(stretchLeft)
				{
					case smSingle:
						imgLeft->draw(Vector2D(rect.pos.x-imgLeft->getWidth(),rect.pos.y));
						break;

					case smRepeat:
						imgLeft->draw(Rect(Vector2D(rect.pos.x-imgLeft->getWidth(),rect.pos.y),Vector2D(imgLeft->getWidth(),rect.dimensions.y)));
						break;

					case smStretch:
						imgLeft->drawStretched(Rect(Vector2D(rect.pos.x-imgLeft->getWidth(),rect.pos.y),Vector2D(imgLeft->getWidth(),rect.dimensions.y)));
						break;
				}
			}

			if(imgCenter)
			{
				switch(stretchCenter)
				{
					case smSingle:
						imgCenter->draw(rect.pos);
						break;

					case smRepeat:
						imgCenter->draw(rect);
						break;

					case smStretch:
						imgCenter->drawStretched(rect);
						break;
				}
			}

			if(imgRight)
			{
				switch(stretchRight)
				{
					case smSingle:
						imgRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y));
						break;

					case smRepeat:
						imgRight->draw(Rect(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y),Vector2D(imgRight->getWidth(),rect.dimensions.y)));
						break;

					case smStretch:
						imgRight->drawStretched(Rect(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y),Vector2D(imgRight->getWidth(),rect.dimensions.y)));
						break;
				}
			}

			if(imgLowerLeft)
			{
				imgLowerLeft->draw(Vector2D(rect.pos.x-imgUpperLeft->getWidth(),rect.pos.y-imgLowerLeft->getHeight()));
			}

			if(imgLower)
			{
				switch(stretchLower)
				{
					case smSingle:
						imgLower->draw(Vector2D(rect.pos.x,rect.pos.y-imgLower->getHeight()));
						break;

					case smRepeat:
						imgLower->draw(Rect(Vector2D(rect.pos.x,rect.pos.y-imgLower->getHeight()),Vector2D(rect.dimensions.x,imgLower->getHeight())));
						break;

					case smStretch:
						imgLower->drawStretched(Rect(Vector2D(rect.pos.x,rect.pos.y-imgLower->getHeight()),Vector2D(rect.dimensions.x,imgLower->getHeight())));
						break;
				}
			}

			if(imgLowerRight)
			{
				imgLowerRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y-imgLowerRight->getHeight()));
			}
			/*			if(true)//stretch)
			{
				if(imgUpperLeft)
				{
					imgUpperLeft->draw(Vector2D(rect.pos.x-imgUpperLeft->getWidth(),rect.pos.y+rect.dimensions.y));
				}

				if(imgUpper)
				{
					imgUpper->drawStretched(Rect(Vector2D(rect.pos.x,rect.pos.y+rect.dimensions.y),
																			 Vector2D(rect.dimensions.x,imgUpper->getHeight())));
				}

				if(imgUpperRight)
				{
					imgUpperRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y));
				}

				if(imgLeft)
				{
					imgLeft->drawStretched(Rect(Vector2D(rect.pos.x-imgLeft->getWidth(),rect.pos.y),
																			Vector2D(imgLeft->getWidth(),rect.dimensions.y)));
				}

				if(imgCenter)
				{
					imgCenter->drawStretched(rect);
				}

				if(imgRight)
				{
					imgRight->drawStretched(Rect(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y),
																			 Vector2D(imgRight->getWidth(),rect.dimensions.y)));
				}

				if(imgLowerLeft)
				{
					imgLowerLeft->draw(Vector2D(rect.pos.x-imgUpperLeft->getWidth(),rect.pos.y-imgLowerLeft->getHeight()));
				}

				if(imgLower)
				{
					imgLower->drawStretched(Rect(Vector2D(rect.pos.x,rect.pos.y-imgLower->getHeight()),
																			 Vector2D(rect.dimensions.x,imgLower->getHeight())));
				}

				if(imgLowerRight)
				{
					imgLowerRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y-imgLowerRight->getHeight()));
				}
			}
			else
			{
				if(imgUpperLeft)
				{
					imgUpperLeft->draw(Vector2D(rect.pos.x-imgUpperLeft->getWidth(),rect.pos.y+rect.dimensions.y));
				}

				if(imgUpper)
				{
					imgUpper->draw(Rect(Vector2D(rect.pos.x,rect.pos.y+rect.dimensions.y),
															Vector2D(rect.dimensions.x,imgUpper->getHeight())));
				}

				if(imgUpperRight)
				{
					imgUpperRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y+rect.dimensions.y));
				}

				if(imgLeft)
				{
					imgLeft->draw(Rect(Vector2D(rect.pos.x-imgLeft->getWidth(),rect.pos.y),
														 Vector2D(imgLeft->getWidth(),rect.dimensions.y)));
				}

				if(imgCenter)
				{
					imgCenter->draw(rect);
				}

				if(imgRight)
				{
					imgRight->draw(Rect(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y),
															Vector2D(imgRight->getWidth(),rect.dimensions.y)));
				}

				if(imgLowerLeft)
				{
					imgLowerLeft->draw(Vector2D(rect.pos.x-imgUpperLeft->getWidth(),rect.pos.y-imgLowerLeft->getHeight()));
				}

				if(imgLower)
				{
					imgLower->draw(Rect(Vector2D(rect.pos.x,rect.pos.y-imgLower->getHeight()),
															Vector2D(rect.dimensions.x,imgLower->getHeight())));
				}

				if(imgLowerRight)
				{
					imgLowerRight->draw(Vector2D(rect.pos.x+rect.dimensions.x,rect.pos.y-imgLowerRight->getHeight()));
				}
		}*/

			glPopAttrib();

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void TexturedRectStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("TexturedRectStyle::getFromNode()");

			RectStyle::getFromNode(node,gui);

			if(const std::string *stretch=node->getAttribute(XML_GUI_STRETCH))
			{
				StretchMode stretchMode=strTo<StretchMode>(*stretch);

				stretchUpper=stretchMode;
				stretchLower=stretchMode;
				stretchLeft=stretchMode;
				stretchRight=stretchMode;
				stretchCenter=stretchMode;
			}

			if(const std::string *stretch=node->getAttribute(XML_GUI_STRETCH_UPPER))
			{
				stretchUpper=strTo<StretchMode>(*stretch);
			}

			if(const std::string *stretch=node->getAttribute(XML_GUI_STRETCH_LOWER))
			{
				stretchLower=strTo<StretchMode>(*stretch);
			}

			if(const std::string *stretch=node->getAttribute(XML_GUI_STRETCH_LEFT))
			{
				stretchLeft=strTo<StretchMode>(*stretch);
			}

			if(const std::string *stretch=node->getAttribute(XML_GUI_STRETCH_RIGHT))
			{
				stretchRight=strTo<StretchMode>(*stretch);
			}

			if(const std::string *stretch=node->getAttribute(XML_GUI_STRETCH_CENTER))
			{
				stretchCenter=strTo<StretchMode>(*stretch);
			}

			if(const std::string *color=node->getAttribute(XML_GUI_COLOR))
			{
				this->color.set((*color).c_str());
			}

			if(const std::string *color=node->getAttribute(XML_GUI_COLORBORDER))
			{
				this->colorBorder.set((*color).c_str());
			}

			if(const std::string *image=node->getAttribute(XML_GUI_CENTER))
			{
				imgCenter=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_UPPERLEFT))
			{
				imgUpperLeft=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_UPPER))
			{
				imgUpper=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_UPPERRIGHT))
			{
				imgUpperRight=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_LEFT))
			{
				imgLeft=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_RIGHT))
			{
				imgRight=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_LOWERLEFT))
			{
				imgLowerLeft=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_LOWER))
			{
				imgLower=(*gui->getImages())[*image];
			}

			if(const std::string *image=node->getAttribute(XML_GUI_LOWERRIGHT))
			{
				imgLowerRight=(*gui->getImages())[*image];
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  LabelStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		LabelStyle::LabelStyle() : Style(),rect(0),font(0),
			imageHorizontal(jhCenter),imageVertical(jvCenter),imageColor(1.0,1.0,1.0,1.0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void LabelStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("LabelStyle::getFromNode()");

			Style::getFromNode(node,gui);

			if(const std::string *style=node->getAttribute(XML_GUI_RECT))
			{
				rect=dynamic_cast<RectStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *style=node->getAttribute(XML_GUI_FONT))
			{
				font=dynamic_cast<FontStyle *>((*gui->getStyles())[*style]);
			}

			imageHorizontal=justificationHorizontal(
											node->getAttribute(XML_GUI_IMAGE_HORIZONTAL,
											STR_JUSTIFICATION_HORIZONTAL[imageHorizontal]).c_str());

			imageVertical  =justificationVertical(
											node->getAttribute(XML_GUI_IMAGE_VERTICAL,
											STR_JUSTIFICATION_VERTICAL[imageVertical]).c_str());

			if(const std::string *color=node->getAttribute(XML_GUI_IMAGECOLOR))
			{
				imageColor.set((*color).c_str());
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void LabelStyle::draw(const Rect &rect)
		{
			//debug::addMessage("LabelStyle::draw()");

			if(this->rect)
			{
				this->rect->draw(rect);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void LabelStyle::print(Rect rect,const char *string,const mysdl::gfx::Image *image)
		{
			//debug::addMessage("LabelStyle::print()");

			if(image)
			{
				Vector2D pos;

				switch(imageHorizontal)
				{
					case jhLeft:
						pos.x=rect.pos.x;

						if(font->justificationHorizontal==jhLeft)
						{
							rect.reduceLeft(image->getWidth());
						}

						break;

					case jhCenter:
						pos.x=rect.pos.x+rect.dimensions.x/2-(GLfloat)(image->getWidth())/2;
						break;

					case jhRight:
						pos.x=rect.pos.x+rect.dimensions.x-image->getWidth();

						if(font->justificationHorizontal==jhRight)
						{
							rect.reduceRight(image->getWidth());
						}

						break;

					default:
						assert(0);
				}

				switch(imageVertical)
				{
					case jvBottom:
						pos.y=rect.pos.y;

						if(font->justificationVertical==jvBottom)
						{
							rect.reduceBottom(image->getHeight());
						}
						break;

					case jvCenter:
						pos.y=rect.pos.y+rect.dimensions.y/2-(GLfloat)(image->getHeight())/2;
						break;

					case jvTop:
						pos.y=rect.pos.y+rect.dimensions.y-image->getHeight();

						if(font->justificationVertical==jvTop)
						{
							rect.reduceTop(image->getHeight());
						}

						break;

					default:
						assert(0);
				}

				imageColor.apply();
				image->draw(pos);
			}

			if(font)
			{
				this->font->print(rect,string);
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  ButtonStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		ButtonStyle::ButtonStyle() : Style(),up(0),down(0),upHot(0),downHot(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void ButtonStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("ButtonStyle::getFromNode()");

			Style::getFromNode(node,gui);

			if(const std::string *style=node->getAttribute(XML_GUI_UP))
			{
				up=dynamic_cast<LabelStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *style=node->getAttribute(XML_GUI_DOWN))
			{
				down=dynamic_cast<LabelStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *style=node->getAttribute(XML_GUI_UPHOT))
			{
				upHot=dynamic_cast<LabelStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *style=node->getAttribute(XML_GUI_DOWNHOT))
			{
				downHot=dynamic_cast<LabelStyle *>((*gui->getStyles())[*style]);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		LabelStyle *ButtonStyle::getStyle(ButtonState state)
		{
			if(state & bsDown)
			{
				return(state & bsHot ? downHot : down);
			}
			else
			{
				return(state & bsHot ? upHot : up);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void ButtonStyle::draw(const Rect &rect,ButtonState state)
		{
			//debug::addMessage("ButtonStyle::draw()");

			LabelStyle *ls=getStyle(state);

			if(ls)
			{
				ls->draw(rect);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void ButtonStyle::print(const Rect &rect,ButtonState state,const char *string,const mysdl::gfx::Image *image)
		{
			//debug::addMessage("ButtonStyle::print()");
			getStyle(state)->print(rect,string,image);
			//debug::removeMessage();
		}

		// ================================================================
		//  CheckBoxStyle
		// ================================================================

		/*
		 * See header file for documentation.
		 */

		CheckboxStyle::CheckboxStyle() : Style(),checked(0),unchecked(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void CheckboxStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("CheckboxStyle::getFromNode()");

			Style::getFromNode(node,gui);

			if(const std::string *style=node->getAttribute(XML_GUI_CHECKED))
			{
				checked=dynamic_cast<ButtonStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *style=node->getAttribute(XML_GUI_UNCHECKED))
			{
				unchecked=dynamic_cast<ButtonStyle *>((*gui->getStyles())[*style]);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		ButtonStyle *CheckboxStyle::getStyle(bool checked)
		{
			return(checked ? this->checked : this->unchecked);
		}

		/*
		 * See header file for documentation.
		 */
		LabelStyle *CheckboxStyle::getStyle(ButtonState state,bool checked)
		{
			return(getStyle(checked)->getStyle(state));
		}

		/*
		 * See header file for documentation.
		 */
		void CheckboxStyle::draw(const Rect &rect,ButtonState state,bool checked)
		{
			//debug::addMessage("CheckboxStyle::draw()");
			getStyle(state,checked)->draw(rect);
			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void CheckboxStyle::print(const Rect &rect,ButtonState state,bool checked,const char *string,const mysdl::gfx::Image *image)
		{
			//debug::addMessage("CheckboxStyle::print()");
			getStyle(state,checked)->print(rect,string,image);
			//debug::removeMessage();
		}

		// ================================================================
		//  DropDownListStyle
		// ================================================================

		DropdownListStyle::DropdownListStyle() : ButtonStyle(),image(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void DropdownListStyle::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("DropdownListStyle::getFromNode()");

			ButtonStyle::getFromNode(node,gui);

			if(const std::string *image=node->getAttribute(XML_GUI_IMAGE))
			{
				this->image=((*gui->getImages())[*image]);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void DropdownListStyle::print(const Rect &rect,ButtonState state,const char *string)
		{
			//debug::addMessage("DropdownListStyle::draw()");
			getStyle(state)->print(rect,string,image);
			//debug::removeMessage();
		}
	};
};
