/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "text.h"
#include "xml.h"

namespace mysdl
{
	namespace gui
	{
		// ================================================================
		//  Label
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Label::Label() : Component(),CaptionedObject(),style(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void Label::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Label::getSelfFromNode()");

			Component::getSelfFromNode(node,gui);
			CaptionedObject::getSelfFromNode(node,gui);

			if(const std::string *style=node->getAttribute(XML_GUI_STYLE))
			{
				this->style=dynamic_cast<LabelStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *image=node->getAttribute(XML_GUI_IMAGE))
			{
				setImage((*gui->getImages())[*image]);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Label::drawSelf()
		{
			//debug::addMessage("Label::drawSelf()");

			if(style)
			{
				mysdl::Rect rect;
				rect.pos=Vector2D(0,0);
				rect.dimensions=dimensions;
				style->draw(rect);

				glPushMatrix();
				style->print(rect,caption.c_str(),image);
				glPopMatrix();
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Label::setStyle(LabelStyle *style)
		{
			this->style=style;
		}

		/*
		 * See header file for documentation.
		 */
		LabelStyle *Label::getStyle()
		{
			return(style);
		}

		// ================================================================
		//  Textfield
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Textfield::Textfield() : Component(),CaptionedObject(),keydown(0),repeatTimer(0),style(0)
		{
			wantsFocus=true;
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Textfield::getSelfFromNode()");

			Component::getSelfFromNode(node,gui);
			CaptionedObject::getSelfFromNode(node,gui);

			if(const std::string *style=node->getAttribute(XML_GUI_STYLE))
			{
				this->style=dynamic_cast<LabelStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *image=node->getAttribute(XML_GUI_IMAGE))
			{
				//this->image=((*gui->getImages())[*image]);
				setImage((*gui->getImages())[*image]);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_CONFIRM))
			{
				this->triggerConfirm=gui->findTrigger(*trigger);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::drawSelf()
		{
			//debug::addMessage("Textfield::drawSelf()");

			if(style)
			{
				mysdl::Rect rect;
				rect.pos=Vector2D(0,0);
				rect.dimensions=dimensions;
				style->draw(rect);

				std::string text=caption;

				if(isFocused())
				{
					text+="_";
				}

				glPushMatrix();
				style->print(rect,text.c_str(),image);
				glPopMatrix();
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::doConfirm(ControlQueue *queue)
		{
			if(triggerConfirm)
			{
				(*queue) << triggerConfirm;
			}
		}

		/*
		 * See header file for documentation.
		 */
		bool Textfield::keyUp(ControlQueue *queue,SDLKey key,SDLMod modifier,Uint16 unicode)
		{
			keydown=0;
			return(true);
		}

		/*
		 * See header file for documentation.
		 */
		bool Textfield::keyDown(ControlQueue *queue,SDLKey key,SDLMod modifier,Uint16 unicode)
		{
			if(unicode==8)//SDLK_BACKSPACE)
			{
				if(caption.size()>0)
				{
					caption.erase(caption.size()-1);

					if(unicode!=keydown)
					{
						keydown=unicode;
						repeatTimer=-getRoot()->getKeyDelay();
					}
				}
				else
				{
					keydown=0;
				}

				return(true);
			}
			else if(unicode==13)
			{
				keydown=0;
				doConfirm(queue);
				return(true);
			}
			else if((unicode>=32)&&(unicode<=255))
			{
				caption+=(char)unicode;//key;

				//if(key!=unicode)
				if(unicode!=keydown)
				{
					keydown=unicode;
					repeatTimer=-getRoot()->getKeyDelay();
				}

				return(true);
			}
			else
			{
				return(Component::keyDown(queue,key,modifier,unicode));
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::enter(ControlQueue *queue,Vector2D pos)
		{
			//SDL_EnableUNICODE(1);
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::leave(ControlQueue *queue,Vector2D pos)
		{
			//SDL_EnableUNICODE(0);
			keydown=0;
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::performSelf(ControlQueue *queue,GLfloat ticks)
		{
			if(keydown)
			{
				GLfloat repeat=getRoot()->getKeyRepeat();

				repeatTimer+=ticks;

				while(repeatTimer>=repeat)
				{
					repeatTimer-=repeat;
					keyDown(queue,SDLK_PAUSE,KMOD_NONE,keydown); // \todo replace SDLK_PAUSE
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::setStyle(LabelStyle *style)
		{
			this->style=style;
		}

		/*
		 * See header file for documentation.
		 */
		LabelStyle *Textfield::getStyle()
		{
			return(style);
		}

		/*
		 * See header file for documentation.
		 */
		void Textfield::setTriggerConfirm(Trigger *trigger)
		{
			triggerConfirm=trigger;
		}

		/*
		 * See header file for documentation.
		 */
		Trigger *Textfield::getTriggerConfirm()
		{
			return(triggerConfirm);
		}
	};
};
