/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <assert.h>
#include <string>
#include <iostream>
#include <fstream>

#include "../mysdl.h"

#include "sprites.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))

namespace mysdl
{
	/*
	 * See header file for documentation.
	 */
	namespace sprites
	{
		// ================================================================
		//  Sprite
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Sprite::Sprite(const Vector2D pos,const bool visible) :
			owner(0),
			color(1,1,1,1),
			pos(pos),hotspot(),visible(visible),
			canCollide(false),blendMode(bmDefault),
			outerBoundingBox(),innerBoundingBox()
		{
		}

		/*
		 * See header file for documentation.
		 */
		Sprite::~Sprite()
		{
			if(owner)
			{
				owner->remove(this);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Sprite::draw()
		{
			if((!owner)||(blendMode!=owner->blendMode))
			{
				glPushAttrib(GL_COLOR);
				setBlendMode(blendMode);
			}

			drawSelf();

			if((!owner)||(blendMode!=owner->blendMode))
			{
				glPopAttrib();
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Sprite::perform(GLfloat ticks)
		{
			// Do nothing
		}

		/*
		 * See header file for documentation.
		 */
		void Sprite::onCollision(Sprite *spr)
		{
			// Do nothing
		}

		/*
		 * See header file for documentation.
		 */
		mysdl::gfx::Bitmask *Sprite::getBitmask() const
		{
			return(0);
		}

		/*
		 * See header file for documentation.
		 */
		bool Sprite::checkCollisionOuterBoundingBox(const Sprite *spr)
		{
			RectInt r1=outerBoundingBox;
			RectInt r2=spr->outerBoundingBox;
			
			r1.pos.x-=round<Sint32>((spr->pos.x-spr->hotspot.x)-(pos.x-hotspot.x));
			r1.pos.y-=round<Sint32>((pos.y+hotspot.y)-(spr->pos.y+spr->hotspot.y));

			return r1.overlap(r2);
		}

		/*
		 * See header file for documentation.
		 */
		bool Sprite::checkCollisionInnerBoundingBox(const Sprite *spr)
		{
			RectInt r1=innerBoundingBox;
			RectInt r2=spr->innerBoundingBox;

			r1.pos.x-=round<Sint32>((spr->pos.x-spr->hotspot.x)-(pos.x-hotspot.x));
			r1.pos.y-=round<Sint32>((pos.y+hotspot.y)-(spr->pos.y+spr->hotspot.y));

			return r1.overlap(r2);
		}

		/*
		 * See header file for documentation.
		 */
		bool Sprite::checkCollisionBitmask(const Sprite *spr) const
		{
			bool b=mysdl::gfx::checkCollisionBitmask(getBitmask(),spr->getBitmask(),
							round<Sint32>((spr->pos.x-spr->hotspot.x)-(pos.x-hotspot.x)),
							round<Sint32>((pos.y+hotspot.y)-(spr->pos.y+spr->hotspot.y)));

			return(b);
		}

		/*
		 * See header file for documentation.
		 */
		bool Sprite::checkCollision(const Sprite *spr)
		{
			if(checkCollisionOuterBoundingBox(spr))
			{
				if(checkCollisionInnerBoundingBox(spr))
				{
					return(true);
				}

				if((getBitmask())&&(spr->getBitmask()))
				{
					return(checkCollisionBitmask(spr));
				}
				else
				{
					return(true);
				}
			}
			else
			{
				return(false);
			}
		}

		// ================================================================
		//  ImageSprite
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		ImageSprite::ImageSprite(Vector2D pos,const bool visible,mysdl::gfx::Image *const image) :
				Sprite(pos,visible),image(0)
		{
			for(int i=0; i<4; i++)
			{
				for(int j=0; j<4; j++)
				{
					vertexColors[i][j]=1;
				}
			}

			setImage(image);
		}

		/*
		 * See header file for documentation.
		 */
		ImageSprite::~ImageSprite()
		{
			autoclearImage();
		}

		/*
		 * See header file for documentation.
		 */
		mysdl::gfx::Bitmask *ImageSprite::getBitmask() const
		{
			return(image->getBitmask());
		}

		/*
		 * See header file for documentation.
		 */
		void ImageSprite::autoclearImage()
		{
			mysdl::gfx::AnimationReference *ref=dynamic_cast<mysdl::gfx::AnimationReference *>(image);

			if(ref)
			{
				delete ref;
			}

			image=0;
		}

		/*
		 * See header file for documentation.
		 */
		mysdl::gfx::AnimationReference *ImageSprite::setAnimation(mysdl::gfx::Animation *const animation,GLfloat delay)
		{
			mysdl::gfx::AnimationReference *ref=animation->getReference(delay);
			setImage(ref);

			return ref;
		}

		/*
		 * See header file for documentation.
		 */
		void ImageSprite::getImageData()
		{
			if(image)
			{
				dimensions.x=image->getWidth();
				dimensions.y=image->getHeight();
				hotspot=image->getHotspot();
				outerBoundingBox=*(image->getOuterBoundingBox());
				innerBoundingBox=*(image->getInnerBoundingBox());
			}
			else
			{
				dimensions.x=dimensions.y=0;
				hotspot.x=hotspot.y=0;
				outerBoundingBox.clear();
				innerBoundingBox.clear();
			}
		}

		/*
		 * See header file for documentation.
		 */
		void ImageSprite::setImage(mysdl::gfx::Image *const image)
		{
			autoclearImage();
			this->image=image;
			getImageData();
		}

		/*
		 * See header file for documentation.
		 */
		void ImageSprite::drawSelf()
		{
			color.apply();
			image->draw(Vector2D(pos.x,pos.y-dimensions.y));
		}

		/*
		 * See header file for documentation.
		 */
		void ImageSprite::perform(GLfloat ticks)
		{
			if(image)
			{
				image->perform(ticks);
				getImageData();
			}
		}

		// ================================================================
		//  SpriteEngine
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		SpriteEngine::SpriteEngine() : sprites(),blendMode(bmDefault)
		{
		}

		/*
		 * See header file for documentation.
		 */
		SpriteEngine::~SpriteEngine()
		{
			clear();
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::draw()
		{
			glPushAttrib(GL_COLOR_BUFFER_BIT);

			glEnable(GL_BLEND);
			setBlendMode(blendMode);

			for(SpriteList::const_iterator i=sprites.begin(); i!=sprites.end(); i++)
			{
				if((*i)->visible)
				{
					(*i)->draw();
				}
			}

			glPopAttrib();
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::perform(GLfloat ticks)
		{
			for(SpriteList::const_iterator i=sprites.begin(); i!=sprites.end(); i++)
			{
				(*i)->perform(ticks);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::testCollisions()
		{
			for(SpriteList::const_iterator i=sprites.begin(); i!=sprites.end(); i++)
			{
				if((*i)->canCollide)
				{
					for(SpriteList::const_iterator i2=i; i2!=sprites.end(); i2++)
					{
						if((*i!=*i2)&&((*i2)->canCollide))
						{
							if((*i)->checkCollision(*i2))
							{
								(*i)->onCollision(*i2);
							}
						}
					}
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::testCollisions(SpriteEngine *e2)
		{
			for(SpriteList::const_iterator i=sprites.begin(); i!=sprites.end(); i++)
			{
				if((*i)->canCollide)
				{
					for(SpriteList::const_iterator i2=e2->sprites.begin(); i2!=e2->sprites.end(); i2++)
					{
						if(((*i2)->canCollide)&&((*i)->checkCollision(*i2)))
						{
							(*i)->onCollision(*i2);
							(*i2)->onCollision(*i);
						}
					}
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::addBack(Sprite *sprite)
		{
			if(sprite->owner)
			{
				sprite->owner->remove(sprite);
			}

			sprites.push_back(sprite);
			sprite->owner=this;
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::addFront(Sprite *sprite)
		{
			if(sprite->owner)
			{
				sprite->owner->remove(sprite);
			}

			sprites.push_front(sprite);
			sprite->owner=this;
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::remove(Sprite *sprite)
		{
			if(sprite->owner==this)
			{
				sprites.remove(sprite);
				sprite->owner=0;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void SpriteEngine::clear()
		{
			while(!(sprites.empty()))
			{
				remove(sprites.front());
			}
		}

		/*
		 * See header file for documentation.
		 */
		int SpriteEngine::size()
		{
			return(sprites.size());
		}
	};
};
