/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <fstream>
#include <iostream>
#include <stdio.h>

#include <expat.h>

#include "debug.h"
#include "mysdl.h"
#include "xml.h"

namespace mysdl
{
	namespace xml
	{
		// ================================================================
		//  Common functions #1
		// ================================================================

		// TODO: switch from iso-8859-1 to UTF-8?

		/**
		 * Encodes a string in a suitable format for XML files by
		 * substituting entities for the XML-specific control characters.
		 * Other characters are left unchanged, so input string should
		 * conform to the ISO-8859-1 standard.
		 *
		 * @param s Input string.
		 *
		 * @returns Encoded string.
		 */
		std::string encode(std::string s)
		{
			std::string res;

			for(std::string::iterator itr=s.begin(); itr!=s.end(); itr++)
			{
				switch(*itr)
				{
					case '<':
						res+="&lt;";
						break;

					case '>':
						res+="&gt;";
						break;

					case '&':
						res+="&amp;";
						break;

					case '"':
						res+="&quot;";
						break;

					default:
						res+=*itr;
				}
			}

			return(res);
		}

		/**
		 * Helper method called the XML parser when a new XML element
		 * starts in a file parsed by XMLReader. <code>data</code> then
		 * points to the XMLHandler associated with the XMLReader, and
		 * the handler's startElement() method is called to handle the
		 * event.
		 */
		void xmlStart(void *data,const char *name,const char **attr)
		{
			////debug::addMessage("xmlStart()");
			((XMLHandler *)data)->startElement(name,attr);
			//debug::removeMessage();
		}

		/**
		 * Helper method called the XML parser when anXML element end
		 * in a file parsed by XMLReader. <code>data</code> then points
		 * to the XMLHandler associated with the XMLReader, and * the
		 * handler's endElement() method is called to handle the event.
		 */
		void xmlEnd(void *data,const char *name)
		{
			////debug::addMessage("xmlEnd()");
			((XMLHandler *)data)->endElement(name);
			//debug::removeMessage();
		}

		// ================================================================
		//  XMLReader
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		XMLReader::XMLReader(XMLHandler *handler) : handler(handler)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void XMLReader::readFile(std::string filename)
		{
			////debug::addMessage("XMLReader::load()");

			FILE *file=fopen(filename.c_str(),"r");

			if(!file)
			{
				//debug::removeMessage();
				throw std::string(std::string("Could not open XML file \"")+filename+std::string("\""));
			}

			static const size_t BUFFERSIZE=1023;

			char buf[BUFFERSIZE+1];
			memset(buf,0,BUFFERSIZE+1);

			XML_Parser parser=XML_ParserCreate(0);
			XML_SetUserData(parser,handler);
			XML_SetElementHandler(parser,xmlStart,xmlEnd);

			while(!feof(file))
			{
				size_t len=fread(buf,1,BUFFERSIZE,file);
				bool done=(len<BUFFERSIZE);

				buf[len]=0;

				if(ferror(file))
				{
					fclose(file);
					XML_ParserFree(parser);
					//debug::removeMessage();
					throw std::string(std::string("Error reading XML file"));
				}
				else if((len>0)&&(XML_Parse(parser,buf,len,done)==XML_STATUS_ERROR))
				{
					std::string ex=XML_ErrorString(XML_GetErrorCode(parser));
					ex+=" at line ";

					fclose(file);
					XML_ParserFree(parser);
					//debug::removeMessage();

					throw ex;
				}
			}

			fclose(file);
			XML_ParserFree(parser);

			//debug::removeMessage();
		}

		// ================================================================
		//  Node
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Node::Node() : parent(0),name(),attributes(),children()
		{
		}

		/*
		 * See header file for documentation.
		 */
		Node::Node(const char *name,const char **attributes) : parent(0),name(name),attributes(),children()
		{
			////debug::addMessage("Node(const char *,const char **)");

			if(attributes)
			{
				for(int i=0; attributes[i]; i+=2)
				{
					this->attributes[attributes[i]]=attributes[i+1];
				}
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		Node::~Node()
		{
			////debug::addMessage("~Node()");

			if(parent)
			{
				(*parent)-=this;
				parent=0;
			}

			while(!children.empty())
			{
				delete children.front();
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Node::operator+=(Node *node)
		{
			////debug::addMessage("Node::operator+=()");

			if(node->parent)
			{
				(*node->parent)-=node;
			}

			children.push_back(node);
			node->parent=this;

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Node::operator-=(Node *node)
		{
			////debug::addMessage("Node::operator-=()");

			children.remove(node);
			node->parent=0; // TODO: check if really owner?

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		std::string &Node::operator[](std::string name)
		{
			return(attributes[name]);
		}

		/*
		 * See header file for documentation.
		 */
		const std::string *Node::getAttribute(const std::string name) const
		{
			Attributes::const_iterator itr=attributes.find(name);
			return(itr==attributes.end() ? 0 : &(itr->second));
		}

		/*
		 * See header file for documentation.
		 */
		std::string Node::getAttribute(const std::string name,const std::string defValue) const
		{
			const std::string *value=getAttribute(name);

			return(value ? *value : defValue);
		}

		/*
		 * See header file for documentation.
		 */
		long Node::getAttributeLong(const std::string name,const long defValue) const
		{
			const std::string *s=getAttribute(name);
			return(s ? strtol(s->c_str(),0,0) : defValue);
		}

		/*
		 * See header file for documentation.
		 */
		void Node::setAttributeLong(const std::string name,const long value)
		{
			(*this)[name]=toStr(value);
		}

		/*
		 * See header file for documentation.
		 */
		float Node::getAttributeFloat(const std::string name,const float defValue) const
		{
			const std::string *s=getAttribute(name);
			return(s ? strtof(s->c_str(),0) : defValue);
		}

		/*
		 * See header file for documentation.
		 */
		void Node::setAttributeFloat(const std::string name,const float value)
		{
			(*this)[name]=toStr(value);
		}

		/*
		 * See header file for documentation.
		 */
		bool Node::getAttributeBool(const std::string name,const bool defValue) const
		{
			const std::string *s=getAttribute(name);
			return(s ? (*s=="1")||(*s=="t")||(*s=="true") : defValue);
		}

		/*
		 * See header file for documentation.
		 */
		void Node::setAttributeBool(const std::string name,const bool value)
		{
			(*this)[name]=(value ? "true" : "false");
		}

		/*
		 * See header file for documentation.
		 */
		std::string Node::toString(int indent) const
		{
			////debug::addMessage("Node::toString()");

			std::string indentStr="";

			for(int i=0; i<indent; i++)
			{
				indentStr+=' ';
			}

			std::string s=indentStr+"<"+name;

			for(Attributes::const_iterator itr=attributes.begin(); itr!=attributes.end(); itr++)
			{
				s+=" "+itr->first+"=\""+encode(itr->second)+"\"";
			}

			if(children.empty())
			{
				s+="/>\n";
			}
			else
			{
				s+=">\n";

				for(Nodes::const_iterator itr=children.begin(); itr!=children.end(); itr++)
				{
					s+=(*itr)->toString(indent+2);
				}

				s+=indentStr+"</"+name+">\n";
			}

			//debug::removeMessage();

			return(s);
		}

		/*
		 * See header file for documentation.
		 */
		void Node::saveXML(const char *filename)
		{
			std::ofstream file(filename);
			file << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" ?>" << std::endl;
			file << toString();
			file.flush();
			file.close();
		}

		// ================================================================
		//  DOMXMLHandler
		// ================================================================

		/**
		 * An XML handler that builds a DOM tree.
		 */
 		class DOMXMLHandler : public XMLHandler
 		{
 			public:
				/** Root node of the DOM tree built by this handler. */
 				Node *root;

				/** Current tree node, depending on the parser's state. */
 				Node *current;

				/**
				 * Creates a new DOMXMLHandler instance.
				 */
				DOMXMLHandler();

				/**
				 * Called when a new XML element begins. The handler will
				 * add this element to the DOM tree.
				 *
				 * @param name  Element name.
				 * @param attrs Element attributes.
				 */
 				void startElement(const char *name,const char **attrs);

				/**
				 * Called when an XML element ends. The handler will
				 * change the current node to represent the new
				 * parser state.
				 *
				 * @param name Element name.
				 */
 				void endElement(const char *name);
 		};

		/*
		 * See interface definition for documentation.
		 */
		DOMXMLHandler::DOMXMLHandler() : root(0),current(0)
		{
		}

		/*
		 * See interface definition for documentation.
		 */
		void DOMXMLHandler::startElement(const char *name,const char **attrs)
		{
			////debug::addMessage("DOMXMLHandler::startElement()");

			Node *node=new Node(name,attrs);

			if(!root)
			{
				root=node;
			}
			else
			{
				(*current)+=node;
			}

			current=node;

			//debug::removeMessage();
		}

		/*
		 * See interface definition for documentation.
		 */
		void DOMXMLHandler::endElement(const char *name)
		{
			////debug::addMessage("DOMXMLHandler::endElement()");
			current=current->parent;
			//debug::removeMessage();
		}

		// ================================================================
		//  LoopData
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		LoopData::LoopData() :
			var(),value(),max(),step(1),digits(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		LoopData::LoopData(std::string var,int value,int max,int step,int digits) :
			var(var),value(value),max(max),step(step),digits(digits)
		{
		}

		/*
		 * See header file for documentation.
		 */
		bool LoopData::advance()
		{
			if(value==max)
			{
				return(false);
			}
			else
			{
				value+=step;
				return(true);
			}
		}

		/*
		 * See header file for documentation.
		 */
		std::string LoopData::getString()
		{
			return(toStr(value,digits,'0'));
		}

		// ================================================================
		//  Common functions #2
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Node *loadXML(const char *filename)
		{
 			////debug::addMessage("loadXML()");

 			DOMXMLHandler *handler=new DOMXMLHandler();
 			XMLReader *reader=new XMLReader(handler);

 			try
 			{
 				reader->readFile(filename);
 			}
 			catch(std::string ex)
 			{
 				delete reader;
 				delete handler;
 				//debug::removeMessage();
 				throw ex;
 			}

			Node *node=handler->root;

 			delete reader;
 			delete handler;

			//debug::removeMessage();

			return(node);
		}
	};
};

